@@ -1,8 +1,9 @@
 {-|
   Copyright   :  (C) 2012-2016, University of Twente,
-                     2016     , Myrtle Software Ltd
+                     2016     , Myrtle Software Ltd,
+                     2021-2022, QBayLogic B.V.
   License     :  BSD2 (see the file LICENSE)
-  Maintainer  :  Christiaan Baaij <christiaan.baaij@gmail.com>
+  Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
   PrettyPrec printing class and instances for CoreHW
 -}
@@ -12,7 +13,6 @@
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE QuasiQuotes #-}
-{-# LANGUAGE TemplateHaskell #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -34,6 +34,7 @@ import Data.Char                        (isSymbol, isUpper, ord)
 import Data.Default                     (Default(..))
 import Data.Text                        (Text)
 import Control.Monad.Identity
+import Data.Binary.IEEE754              (wordToDouble, wordToFloat)
 import Data.List.Extra                  ((<:>))
 import qualified Data.Text              as T
 import Data.Maybe                       (fromMaybe)
@@ -45,7 +46,7 @@ import Data.Text.Prettyprint.Doc
 import Data.Text.Prettyprint.Doc.Internal
 #endif
 import GHC.Show                         (showMultiLineString)
-import Numeric                          (fromRat)
+import GHC.Stack                        (HasCallStack)
 #if MIN_VERSION_ghc(9,0,0)
 import qualified GHC.Utils.Outputable   as GHC
 #else
@@ -59,10 +60,10 @@ import Clash.Core.DataCon               (DataCon (..))
 import Clash.Core.Literal               (Literal (..))
 import Clash.Core.Name                  (Name (..))
 import Clash.Core.Term
-  (Pat (..), Term (..), TickInfo (..), NameMod (..), CoreContext (..), primArg, PrimInfo(primName))
-import Clash.Core.TyCon                 (TyCon (..), TyConName, isTupleTyConLike)
+  (Pat (..), Term (..), TickInfo (..), NameMod (..), CoreContext (..), primArg, PrimInfo(primName),Bind(..))
+import Clash.Core.TyCon                 (TyCon (..), TyConName, isTupleTyConLike, AlgTyConRhs(..))
 import Clash.Core.Type                  (ConstTy (..), Kind, LitTy (..),
-                                         Type (..), TypeView (..), tyView)
+                                         Type (..), TypeView (..), tyView,mkTyConApp)
 import Clash.Core.Var                   (Id, TyVar, Var (..), IdScope(..))
 import Clash.Debug                      (trace)
 import Clash.Util
@@ -219,11 +220,27 @@ pprTopLevelBndr (bndr,expr) = do
   expr'    <- pprM expr
   return $ bndr' <> line <> hang 2 (sep [(bndrName <+> equals), expr']) <> line
 
-dcolon, rarrow, lam, tylam, at, cast, coerce, letrec, in_, case_, of_, forall_
+dcolon, rarrow, lam, tylam, at, cast, coerce, let_, letrec, in_, case_, of_, forall_,
+  data_,newtype_,type_,family_,instance_
   :: ClashDoc
-[dcolon, rarrow, lam, tylam, at, cast, coerce, letrec, in_, case_, of_, forall_]
-  = annotate (AnnSyntax Keyword) <$>
-    ["::", "->", "λ", "Λ", "@", "▷", "~", "letrec", "in", "case", "of", "forall"]
+dcolon = annotate (AnnSyntax Keyword) "::"
+rarrow = annotate (AnnSyntax Keyword) "->"
+lam = annotate (AnnSyntax Keyword)  "λ"
+tylam = annotate (AnnSyntax Keyword) "Λ"
+at = annotate (AnnSyntax Keyword)  "@"
+cast = annotate (AnnSyntax Keyword) "▷"
+coerce = annotate (AnnSyntax Keyword) "~"
+let_ = annotate (AnnSyntax Keyword) "let"
+letrec = annotate (AnnSyntax Keyword) "letrec"
+in_ = annotate (AnnSyntax Keyword) "in"
+case_ = annotate (AnnSyntax Keyword) "case"
+of_ = annotate (AnnSyntax Keyword) "of"
+forall_ = annotate (AnnSyntax Keyword) "forall"
+data_ = annotate (AnnSyntax Keyword) "data"
+newtype_ = annotate (AnnSyntax Keyword) "newtype"
+type_ = annotate (AnnSyntax Keyword) "type"
+family_ = annotate (AnnSyntax Keyword) "family"
+instance_ = annotate (AnnSyntax Keyword) "instance"
 
 instance PrettyPrec Text where
   pprPrec _ = pure . pretty
@@ -235,11 +252,58 @@ instance ClashPretty Type where
   clashPretty = fromPpr
 
 instance PrettyPrec TyCon where
-  pprPrec _ t = pprM (tyConName t)
+  pprPrec prec t = case t of
+    AlgTyCon _ nm kn _ (DataTyCon dcs) _ -> do
+      name <- pprPrec prec nm
+      kind <- pprKind kn
+      let decl = name <> annotate (AnnSyntax Type) (space <> dcolon <+> kind)
+
+      cons <- traverse pprDataCon dcs
+      pure (vsep (data_ <+> decl : cons))
+     where
+      pprDataCon dc = do
+        name <- pprPrec prec dc
+        ty <- pprType (dcType dc)
+
+        pure (name <+> dcolon <+> ty)
+
+    AlgTyCon _ nm kn _ (NewTyCon dc _) _ -> do
+      name <- pprPrec prec nm
+      kind <- pprKind kn
+      let decl = name <> annotate (AnnSyntax Type) (space <> dcolon <+> kind)
+
+      conName <- pprPrec prec (dcName dc)
+      conType <- pprType (dcType dc)
+
+      pure (vsep [newtype_ <+> decl, conName <+> dcolon <+> conType])
+
+    PromotedDataCon _ _ _ _ dc ->
+      fmap ("promoted" <+>) (pprPrec prec dc)
+
+    FunTyCon _ nm kn _ ss -> do
+      name <- pprPrec prec nm
+      kind <- pprKind kn
+      let decl = name <> annotate (AnnSyntax Type) (space <> dcolon <+> kind)
+
+      substs <- traverse pprSubst ss
+      pure (vsep (type_ <+> family_ <+> decl : substs))
+     where
+      pprSubst (xs, y) = do
+        lhs <- pprType (mkTyConApp (tyConName t) xs)
+        rhs <- pprType y
+
+        pure (type_ <+> instance_ <+> lhs <+> "=" <+> rhs)
+
+    PrimTyCon _ nm kn _ -> do
+      name <- pprPrec prec nm
+      kind <- pprKind kn
+
+      pure (name <> annotate (AnnSyntax Type) (space <> dcolon <+> kind))
 
 instance Pretty LitTy where
   pretty (NumTy i) = pretty i
   pretty (SymTy s) = dquotes $ pretty s
+  pretty (CharTy c) = squotes $ pretty c
 
 instance PrettyPrec LitTy where
   pprPrec _ = return . annotate (AnnSyntax LitS) . pretty
@@ -260,7 +324,8 @@ instance PrettyPrec Term where
     App fun arg     -> pprPrecApp prec fun arg
     TyApp e' ty     -> annotate (AnnContext TyAppC) <$>
                          pprPrecTyApp prec e' ty
-    Letrec xes e1   -> pprPrecLetrec prec xes e1
+    Let (NonRec i x) e1 -> pprPrecLetrec prec False [(i,x)] e1
+    Let (Rec xes) e1   -> pprPrecLetrec prec True xes e1
     Case e' _ alts  -> pprPrecCase prec e' alts
     Cast e' ty1 ty2 -> pprPrecCast prec e' ty1 ty2
     Tick t e'       -> do
@@ -276,6 +341,7 @@ instance PrettyPrec TickInfo where
   pprPrec prec (NameMod SetName t)    = ("<setName>" <>) <$> pprPrec prec t
   pprPrec _    DeDup                  = pure "<deDup>"
   pprPrec _    NoDeDup                = pure "<noDeDup>"
+  pprPrec prec (Attributes _ attrs)   = ("<attributes>" <>) <$> pprPrec prec attrs
 
 instance PrettyPrec SrcSpan where
   pprPrec _ sp = return ("<src>"<>pretty (GHC.showSDocUnsafe (GHC.ppr sp)))
@@ -300,20 +366,22 @@ instance PrettyPrec DataCon where
 
 instance PrettyPrec Literal where
   pprPrec _ l = return $ annotate (AnnSyntax LitS) $ case l of
-    IntegerLiteral i
-      | i < 0          -> parens (pretty i)
-      | otherwise      -> pretty i
-    IntLiteral i
-      | i < 0          -> parens (pretty i)
-      | otherwise      -> pretty i
-    Int64Literal i
-      | i < 0          -> parens (pretty i)
-      | otherwise      -> pretty i
-    WordLiteral w      -> pretty w
-    Word64Literal w    -> pretty w
-    FloatLiteral r     -> pretty (fromRat r :: Float)
-    DoubleLiteral r    -> pretty (fromRat r :: Double)
-    CharLiteral c      -> pretty c
+    IntegerLiteral i   -> parensIf (i < 0) (pretty i)
+    IntLiteral i       -> parensIf (i < 0) (pretty i <> "#")
+    Int64Literal i     -> parensIf (i < 0) (pretty i <> "#64")
+    WordLiteral w      -> pretty w <> "##"
+    Word64Literal w    -> pretty w <> "##64"
+#if MIN_VERSION_ghc(8,8,0)
+    Int8Literal i      -> parensIf (i < 0) (pretty i <> "#8")
+    Int16Literal i     -> parensIf (i < 0) (pretty i <> "#16")
+    Int32Literal i     -> parensIf (i < 0) (pretty i <> "#32")
+    Word8Literal w     -> pretty w <> "##8"
+    Word16Literal w    -> pretty w <> "##16"
+    Word32Literal w    -> pretty w <> "##32"
+#endif
+    FloatLiteral w     -> pretty (wordToFloat w) <> "#"
+    DoubleLiteral w    -> pretty (wordToDouble w) <> "##"
+    CharLiteral c      -> pretty c <> "#"
     StringLiteral s    -> vcat $ map pretty $ showMultiLineString s
     NaturalLiteral n   -> pretty n
     ByteArrayLiteral s -> pretty $ show s
@@ -379,10 +447,16 @@ pprPrecCast prec e ty1 ty2 = do
     e' <> annotate (AnnSyntax Type)
                    (softline <> nest 2 (vsep [cast, ty1', coerce, ty2']))
 
-pprPrecLetrec :: Monad m => Rational -> [(Id, Term)] -> Term -> m ClashDoc
-pprPrecLetrec prec xes body = do
+-- TODO Since Clash now keeps non-recursive let expressions separately, the
+-- result of normalization will contain more nested let expressions as the old
+-- Letrec-based definitions are replaced by Let. As this happens, it may be a
+-- good idea to change pprPrecLetrec to encourage more compact forms such as
+-- printing the entire binding on one line if possible.
+
+pprPrecLetrec :: Monad m => Rational -> Bool -> [(Id, Term)] -> Term -> m ClashDoc
+pprPrecLetrec prec isRec xes body = do
   let bndrs = fst <$> xes
-  body' <- annotate (AnnContext $ LetBody bndrs) <$> pprPrec noPrec body
+  body' <- annotate (AnnContext $ LetBody xes) <$> pprPrec noPrec body
   xes'  <- mapM (\(x,e) -> do
                   x' <- pprBndr LetBind x
                   e' <- pprPrec noPrec e
@@ -390,8 +464,9 @@ pprPrecLetrec prec xes body = do
                     vsepHard [x', equals <+> e']
                 ) xes
   let xes'' = case xes' of { [] -> ["EmptyLetrec"]; _  -> xes' }
+  let kw = if isRec then letrec else let_
   return $ parensIf (prec > noPrec) $
-    vsepHard [hang 2 (vsepHard $ letrec : xes''), in_ <+> body']
+    vsepHard [hang 2 (vsepHard $ kw : xes''), in_ <+> body']
 
 pprPrecCase :: Monad m => Rational -> Term -> [(Pat,Term)] -> m ClashDoc
 pprPrecCase prec e alts = do
