@@ -1,8 +1,9 @@
 {-|
   Copyright   :  (C) 2012-2016, University of Twente,
                           2017, Google Inc.
+                          2021, QBayLogic B.V.
   License     :  BSD2 (see the file LICENSE)
-  Maintainer  :  Christiaan Baaij <christiaan.baaij@gmail.com>
+  Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
   Capture-free substitution function for CoreHW
 -}
@@ -10,6 +11,8 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -53,6 +56,10 @@ module Clash.Core.Subst
     -- * Alpha equivalence
   , aeqType
   , aeqTerm
+  , acmpTerm
+    -- * Structural equivalence
+  , eqTerm
+  , eqType
   )
 where
 
@@ -64,20 +71,24 @@ import           Prettyprinter
 import           Data.Text.Prettyprint.Doc
 #endif
 
+import           Data.Hashable             (Hashable (hashWithSalt))
 import qualified Data.List                 as List
 import qualified Data.List.Extra           as List
 import           Data.Ord                  (comparing)
+import           GHC.Stack                 (HasCallStack)
+import           GHC.SrcLoc.Extra          ()
+import           GHC.TypeLits
+  (TypeError, ErrorMessage (Text, (:<>:)))
 
-import           Clash.Core.FreeVars
-  (noFreeVarsOfType, localFVsOfTerms, tyFVsOfTypes)
+import           Clash.Core.HasFreeVars
 import           Clash.Core.Pretty         (ppr, fromPpr)
 import           Clash.Core.Term
-  (LetBinding, Pat (..), Term (..), TickInfo (..), PrimInfo(primName))
+  (Bind(..), Pat (..), Term (..), TickInfo (..), PrimInfo(primName))
 import           Clash.Core.Type           (Type (..))
 import           Clash.Core.VarEnv
 import           Clash.Core.Var            (Id, Var (..), TyVar, isGlobalId)
+import qualified Clash.Data.UniqMap as UniqMap
 import           Clash.Debug               (debugIsOn)
-import           Clash.Unique
 import           Clash.Util
 import           Clash.Pretty
 
@@ -254,7 +265,7 @@ zipTvSubst tvs tys
   , not (List.equalLength tvs tys)
   = pprTrace "zipTvSubst" (ppr tvs <> line <> ppr tys) emptySubst
   | otherwise
-  = Subst (mkInScopeSet (tyFVsOfTypes tys)) emptyVarEnv tenv emptyVarEnv
+  = Subst (mkInScopeSet (freeVarsOf tys)) emptyVarEnv tenv emptyVarEnv
  where
   tenv = zipTyEnv tvs tys
 
@@ -438,7 +449,7 @@ checkValidSubst subst@(TvSubst inScope tenv) tys a =
   WARN( not (isValidSubst subst),
         "inScope" <+> clashPretty inScope <> line <>
         "tenv" <+> clashPretty tenv <> line <>
-        "tenvFVs" <+> clashPretty (tyFVsOfTypes tenv) <> line <>
+        "tenvFVs" <+> clashPretty (freeVarsOf tenv) <> line <>
         "tys" <+> fromPpr tys)
   WARN( not tysFVsInSope,
        "inScope" <+> clashPretty inScope <> line <>
@@ -447,8 +458,8 @@ checkValidSubst subst@(TvSubst inScope tenv) tys a =
        "needsInScope" <+> clashPretty needsInScope)
   a
  where
-  needsInScope = foldrWithUnique (\k _ s -> delVarSetByKey k s)
-                   (tyFVsOfTypes tys)
+  needsInScope = UniqMap.foldrWithUnique (\k _ s -> delVarSetByKey k s)
+                   (freeVarsOf tys)
                    tenv
   tysFVsInSope = needsInScope `varSetInScope` inScope
 
@@ -462,7 +473,7 @@ isValidSubst
   -> Bool
 isValidSubst (TvSubst inScope tenv) = tenvFVs `varSetInScope` inScope
  where
-  tenvFVs = tyFVsOfTypes tenv
+  tenvFVs = freeVarsOf tenv
 
 -- | The work-horse of 'substTy'
 substTy'
@@ -507,11 +518,11 @@ substTyVarBndr subst@(TvSubst inScope tenv) oldVar =
          | otherwise = extendVarEnv oldVar (VarTy newVar) tenv
 
   -- Assertion that we're not capturing something in the substitution
-  no_capture = not (newVar `elemVarSet` tyFVsOfTypes tenv)
+  no_capture = not (newVar `elemVarSet` freeVarsOf tenv)
 
   oldKi        = varType oldVar
   -- verify that the kind is closed
-  noKindChange = noFreeVarsOfType oldKi
+  noKindChange = isClosed oldKi
   -- noChange means that the new type variable is identical in all respects to
   -- the old type variable (same unique, same kind)
   -- See 'TvSubstEnv's Note [Extending the TvSubstEnv]
@@ -557,8 +568,8 @@ substTm doc subst = go where
       (subst',v') -> TyLam v' (substTm doc subst' e)
     App l r -> App (go l) (go r)
     TyApp l r -> TyApp (go l) (substTy subst r)
-    Letrec bs e -> case substBind doc subst bs of
-      (subst',bs') -> Letrec bs' (substTm doc subst' e)
+    Let bs e -> case substBind doc subst bs of
+      (subst',bs') -> Let bs' (substTm doc subst' e)
     Case subj ty alts -> Case (go subj) (substTy subst ty) (map goAlt alts)
     Cast e t1 t2 -> Cast (go e) (substTy subst t1) (substTy subst t2)
     Tick tick e -> Tick (goTick tick) (go e)
@@ -574,6 +585,7 @@ substTm doc subst = go where
   goTick (NameMod m ty) = NameMod m (substTy subst ty)
   goTick t@DeDup        = t
   goTick t@NoDeDup      = t
+  goTick (Attributes ty tm) = Attributes (substTy subst ty) (go tm)
 
 -- | Substitute within a case-alternative
 substAlt
@@ -637,7 +649,7 @@ substIdBndr subst@(Subst inScope env tenv genv) oldId =
         | otherwise    = id1 {varType = substTy subst (varType id1)}
 
   oldTy = varType oldId
-  noTypeChange = nullVarEnv tenv || noFreeVarsOfType oldTy
+  noTypeChange = nullVarEnv tenv || isClosed oldTy
 
   -- Extend the substitution if the unique has changed.
   --
@@ -670,10 +682,16 @@ substBind
   :: HasCallStack
   => Doc ()
   -> Subst
-  -> [LetBinding]
-  -> (Subst,[LetBinding])
-substBind doc subst xs =
-  (subst',zip bndrs' rhss')
+  -> Bind Term
+  -> (Subst, Bind Term)
+substBind doc subst (NonRec i x) =
+  (subst', NonRec i' x')
+ where
+  (subst', i') = substIdBndr subst i
+  x' = substTm ("substBind" <+> doc) subst x
+
+substBind doc subst (Rec xs) =
+  (subst', Rec (zip bndrs' rhss'))
  where
   (bndrs,rhss)    = unzip xs
   (subst',bndrs') = List.mapAccumL substIdBndr subst bndrs
@@ -717,11 +735,11 @@ deshadowLetExpr
   :: HasCallStack
   => InScopeSet
   -- ^ Current InScopeSet
-  -> [LetBinding]
+  -> Bind Term
   -- ^ Bindings of the let-expression
   -> Term
   -- ^ The body of the let-expression
-  -> ([LetBinding],Term)
+  -> (Bind Term, Term)
   -- ^ Deshadowed let-bindings, where let-bound expressions and the let-body
   -- properly reference the renamed variables
 deshadowLetExpr is bs e =
@@ -752,9 +770,9 @@ freshenTm is0 = go (mkSubst is0) where
         (is2,r') -> (is2, App l' r')
     TyApp l r -> case go subst0 l of
       (is1,l') -> (is1, TyApp l' (substTy subst0 r))
-    Letrec bs e -> case goBind subst0 bs of
+    Let bs e -> case goBind subst0 bs of
       (subst1,bs') -> case go subst1 e of
-        (is2,e') -> (is2,Letrec bs' e')
+        (is2,e') -> (is2,Let bs' e')
     Case subj ty alts -> case go subst0 subj of
       (is1,subj') -> case List.mapAccumL (\isN -> goAlt subst0 {substInScope = isN}) is1 alts of
         (is2,alts') -> (is2, Case subj' (substTy subst0 ty) alts')
@@ -764,13 +782,18 @@ freshenTm is0 = go (mkSubst is0) where
        (is1, e') -> (is1, Tick (goTick subst0 tick) e')
     tm -> (substInScope subst0, tm)
 
-  goBind subst0 xs =
+  goBind subst0 (NonRec i x) =
+    let (subst1, i') = substIdBndr subst0 i
+        (is2, x') = go subst0 x
+     in (subst1 { substInScope = extendInScopeSet is2 i' }, NonRec i' x')
+
+  goBind subst0 (Rec xs) =
     let (bndrs,rhss)    = unzip xs
         (subst1,bndrs') = List.mapAccumL substIdBndr subst0 bndrs
         (is2,rhss')     = List.mapAccumL (\isN -> go subst1 {substInScope = isN})
                                          (substInScope subst1)
                                          rhss
-    in  (subst1 {substInScope = is2},zip bndrs' rhss')
+    in  (subst1 {substInScope = is2}, Rec $ zip bndrs' rhss')
 
   goAlt subst0 (pat,alt) = case pat of
     DataPat dc tvs ids -> case List.mapAccumL substTyVarBndr' subst0 tvs of
@@ -792,7 +815,7 @@ aeqType
   -> Bool
 aeqType t1 t2 = acmpType' rnEnv t1 t2 == EQ
  where
-  rnEnv = mkRnEnv (mkInScopeSet (tyFVsOfTypes [t1,t2]))
+  rnEnv = mkRnEnv (mkInScopeSet (freeVarsOf [t1,t2]))
 
 -- | Alpha comparison for types
 acmpType
@@ -801,7 +824,7 @@ acmpType
   -> Ordering
 acmpType t1 t2 = acmpType' (mkRnEnv inScope) t1 t2
  where
-  inScope = mkInScopeSet (tyFVsOfTypes [t1,t2])
+  inScope = mkInScopeSet (freeVarsOf [t1,t2])
 
 -- | Alpha comparison for types. Faster than 'acmpType' as it doesn't need to
 -- calculate the free variables to create the 'InScopeSet'
@@ -819,12 +842,8 @@ acmpType' = go
   go env (AppTy s1 t1) (AppTy s2 t2) =
     go env s1 s2 `thenCompare` go env t1 t2
   go _ (LitTy l1) (LitTy l2) = compare l1 l2
-  go env (AnnType _ t1) (AnnType _ t2) =
-    -- XXX: maybe ignore annotations, like we ignore ticks, i.e.
-    --
-    -- go env (AnnType t1) t2 = go env t1 t2
-    -- go env t1 (AnnType t2) = go env t1 t2
-    go env t1 t2
+  go env (AnnType a1 t1) (AnnType a2 t2) =
+    compare a1 a2 `thenCompare` go env t1 t2
   go _ t1 t2 = compare (getRank t1) (getRank t2)
 
   getRank :: Type -> Word
@@ -835,6 +854,22 @@ acmpType' = go
   getRank (AppTy {})    = 4
   getRank (ForAllTy {}) = 5
 
+-- | Structural equality on 'Type'
+eqType
+  :: Type
+  -> Type
+  -> Bool
+eqType = go
+ where
+  go (VarTy tv1) (VarTy tv2) = tv1 == tv2
+  go (ConstTy c1) (ConstTy c2) = c1 == c2
+  go (ForAllTy tv1 t1) (ForAllTy tv2 t2) =
+    tv1 == tv2 && go (varType tv1) (varType tv2) && go t1 t2
+  go (AppTy s1 t1) (AppTy s2 t2) = go s1 s2 && go t1 t2
+  go (LitTy l1) (LitTy l2) = l1 == l2
+  go (AnnType a1 t1) (AnnType a2 t2) = a1 == a2 && go t1 t2
+  go _ _ = False
+
 -- | Alpha equality for terms
 aeqTerm
   :: Term
@@ -842,7 +877,7 @@ aeqTerm
   -> Bool
 aeqTerm t1 t2 = aeqTerm' inScope t1 t2
  where
-  inScope = mkInScopeSet (localFVsOfTerms [t1,t2])
+  inScope = mkInScopeSet (freeVarsOf [t1,t2])
 
 -- | Alpha equality for terms. Faster than 'aeqTerm' as it doesn't need to
 -- calculate the free variables to create the 'InScopeSet'
@@ -861,7 +896,7 @@ acmpTerm
   -> Ordering
 acmpTerm t1 t2 = acmpTerm' inScope t1 t2
  where
-  inScope = mkInScopeSet (localFVsOfTerms [t1,t2])
+  inScope = mkInScopeSet (freeVarsOf [t1,t2])
 
 -- | Alpha comparison for types. Faster than 'acmpTerm' as it doesn't need to
 -- calculate the free variables to create the 'InScopeSet'
@@ -890,7 +925,9 @@ acmpTerm' inScope = go (mkRnEnv inScope)
     go env l1 l2 `thenCompare` go env r1 r2
   go env (TyApp l1 r1) (TyApp l2 r2) =
     go env l1 l2 `thenCompare` acmpType' env r1 r2
-  go env (Letrec bs1 e1) (Letrec bs2 e2) =
+  go env (Let (NonRec i1 x1) e1) (Let (NonRec i2 x2) e2) =
+    go env x1 x2 `thenCompare` go (rnTmBndr env i1 i2) e1 e2
+  go env (Let (Rec bs1) e1) (Let (Rec bs2) e2) =
     compare (length bs1) (length bs2) `thenCompare`
     foldr thenCmpTm EQ (zipWith (go env') rhs1 rhs2) `thenCompare`
     go env' e1 e2
@@ -906,10 +943,8 @@ acmpTerm' inScope = go (mkRnEnv inScope)
     go env e1 e2 `thenCompare`
     acmpType' env l1 l2 `thenCompare`
     acmpType' env r1 r2
-  -- Look through ticks for aeq
-  go env (Tick _ e1) e2 = go env e1 e2
-  go env e1 (Tick _ e2) = go env e1 e2
-
+  go env (Tick t1 e1) (Tick t2 e2) =
+    compare t1 t2 `thenCompare` go env e1 e2
   go _ e1 e2 = compare (getRank e1) (getRank e2)
 
   goAlt env (DataPat c1 tvs1 ids1,e1) (DataPat c2 tvs2 ids2,e2) =
@@ -930,13 +965,59 @@ acmpTerm' inScope = go (mkRnEnv inScope)
     TyApp {}   -> 6
     Lam {}     -> 7
     TyLam {}   -> 8
-    Letrec {}  -> 9
-    Case {}    -> 10
-    Tick {}    -> 11
+    Let NonRec{} _ -> 9
+    Let Rec{} _ -> 10
+    Case {}    -> 11
+    Tick {}    -> 12
 
-thenCompare :: Ordering -> Ordering -> Ordering
-thenCompare EQ rel = rel
-thenCompare rel _  = rel
+-- | Structural equality on 'Term'
+eqTerm :: Term -> Term -> Bool
+eqTerm = go
+ where
+  go (Var id1) (Var id2) = id1 == id2
+  go (Data dc1) (Data dc2) = dc1 == dc2
+  go (Literal l1) (Literal l2) = l1 == l2
+  go (Prim p1) (Prim p2) = primName p1 == primName p2
+  go (Lam b1 e1) (Lam b2 e2) =
+    b1 == b2 && eqType (varType b1) (varType b2) && go e1 e2
+  go (TyLam b1 e1) (TyLam b2 e2) =
+    b1 == b2 && eqType (varType b1) (varType b2) && go e1 e2
+  go (App l1 r1) (App l2 r2) = go l1 l2 && go r1 r2
+  go (TyApp l1 r1) (TyApp l2 r2) = go l1 l2 && eqType r1 r2
+  go (Let bs1 e1) (Let bs2 e2) =
+    go e1 e2 &&
+    goBind bs1 bs2
+   where
+    goBind (NonRec b1 r1) (NonRec b2 r2) =
+      -- No need to check types of NonRec bindings, when the RHSs match the
+      -- types must be the same.
+      b1 == b2 && go r1 r2
+    goBind (Rec brs1) (Rec brs2) =
+      List.all2
+        (\(b1,r1) (b2,r2) ->
+          b1 == b2 &&
+          -- We need to check the types of Rec bindings, because:
+          --
+          -- letrec (x : Bool) = x in X
+          --
+          -- is not structurally equivalent to
+          --
+          -- letrec (x : Int) = x in x
+          eqType (varType b1) (varType b2) &&
+          go r1 r2)
+        brs1 brs2
+    goBind _ _ = False
+  go (Case e1 _ a1) (Case e2 _ a2) =
+    go e1 e2 &&
+    List.all2 goAlt a1 a2
+   where
+    goAlt (p1,r1) (p2,r2) = p1 == p2 && go r1 r2
+  go (Cast e1 l1 r1) (Cast e2 l2 r2) =
+    go e1 e2 &&
+    eqType l1 l2 &&
+    eqType r1 r2
+  go (Tick t1 e1) (Tick t2 e2) = t1 == t2 && go e1 e2
+  go _ _ = False
 
 instance Eq Type where
   (==) = aeqType
@@ -947,5 +1028,13 @@ instance Ord Type where
 instance Eq Term where
   (==) = aeqTerm
 
+instance TypeError (
+        'Text "A broken implementation of Hashable Term has been "
+  ':<>: 'Text "removed in Clash 1.4.7. If this is an issue for you, please submit "
+  ':<>: 'Text "an issue report at https://github.com/clash-lang/clash-compiler/issues."
+  ) => Hashable Term where
+    hashWithSalt = error "Term.hashWithSalt: unreachable"
+
+
 instance Ord Term where
   compare = acmpTerm
