@@ -1,8 +1,9 @@
 {-|
   Copyright   :  (C) 2012-2016, University of Twente,
                           2017, Google Inc.
+                     2021-2024, QBayLogic B.V.
   License     :  BSD2 (see the file LICENSE)
-  Maintainer  :  Christiaan Baaij <christiaan.baaij@gmail.com>
+  Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
   Term representation in the CoreHW language: System F + LetRec + Case
 -}
@@ -12,10 +13,12 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE PatternSynonyms #-}
 {-# LANGUAGE TemplateHaskell #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 module Clash.Core.Term
-  ( Term (..)
+  ( Term (.., Letrec)
   , mkAbstraction
   , mkTyLams
   , mkLams
@@ -24,8 +27,8 @@ module Clash.Core.Term
   , mkTmApps
   , mkTicks
   , TmName
-  , idToVar
   , varToId
+  , Bind(..)
   , LetBinding
   , Pat (..)
   , patIds
@@ -33,9 +36,11 @@ module Clash.Core.Term
   , Alt
   , TickInfo (..)
   , stripTicks
+  , stripAllTicks
   , partitionTicks
   , NameMod (..)
   , PrimInfo (..)
+  , PrimUnfolding (..)
   , IsMultiPrim (..)
   , MultiPrimInfo (..)
   , WorkInfo (..)
@@ -58,26 +63,28 @@ import Data.Binary                             (Binary)
 import Data.Coerce                             (coerce)
 import qualified Data.DList                    as DList
 import Data.Either                             (lefts, rights)
+#if !MIN_VERSION_base(4,20,0)
 import Data.Foldable                           (foldl')
-import Data.Maybe                              (catMaybes)
+#endif
 import Data.Hashable                           (Hashable)
+import Data.Maybe                              (catMaybes)
 import Data.List                               (nub, partition)
 import Data.Text                               (Text)
 import GHC.Generics
 #if MIN_VERSION_ghc(9,0,0)
-import GHC.Types.SrcLoc                        (SrcSpan)
+import GHC.Types.SrcLoc                        (SrcSpan, leftmost_smallest)
 #else
-import SrcLoc                                  (SrcSpan)
+import SrcLoc                                  (SrcSpan, leftmost_smallest)
 #endif
 
 -- Internal Modules
 import Clash.Core.DataCon                      (DataCon)
 import Clash.Core.Literal                      (Literal)
 import Clash.Core.Name                         (Name (..))
-import {-# SOURCE #-} Clash.Core.Subst         () -- instance Eq Type
+import {-# SOURCE #-} Clash.Core.Subst         (acmpTerm) -- instance Eq/Ord Type, Eq Term
 import {-# SOURCE #-} Clash.Core.Type          (Type)
-import Clash.Core.Var                          (Var(Id), Id, TyVar)
-import Clash.Util                              (curLoc)
+import Clash.Core.Var                          (Var, Id, TyVar)
+import Clash.Util                              (curLoc, thenCompare)
 
 -- | Term representation in the CoreHW language: System F + LetRec + Case
 data Term
@@ -89,12 +96,22 @@ data Term
   | TyLam   !TyVar Term                     -- ^ Type-abstraction
   | App     !Term !Term                     -- ^ Application
   | TyApp   !Term !Type                     -- ^ Type-application
-  | Letrec  [LetBinding] Term               -- ^ Recursive let-binding
+  | Let     !(Bind Term) Term               -- ^ Recursive let-binding
   | Case    !Term !Type [Alt]               -- ^ Case-expression: subject, type of
                                             -- alternatives, list of alternatives
   | Cast    !Term !Type !Type               -- ^ Cast a term from one type to another
   | Tick    !TickInfo !Term                 -- ^ Annotated term
-  deriving (Show,Generic,NFData,Hashable,Binary)
+  deriving (Show, Generic, NFData, Binary)
+
+-- TODO When it is possible, remove this pattern.
+pattern Letrec :: [LetBinding] -> Term -> Term
+pattern Letrec bs x <- Let (bindToList -> bs) x
+ where
+  Letrec bs x = Let (Rec bs) x
+
+bindToList :: Bind a -> [(Id, a)]
+bindToList (NonRec i x) = [(i, x)]
+bindToList (Rec xs) = xs
 
 data TickInfo
   = SrcSpan !SrcSpan
@@ -107,7 +124,25 @@ data TickInfo
   | NoDeDup
   -- ^ Do not deduplicate, i.e. /keep/, an expression inside a case-alternative;
   -- do not try to share expressions between multiple branches.
-  deriving (Eq,Show,Generic,NFData,Hashable,Binary)
+  | Attributes Type Term
+  -- ^ Synthesis attributes brough into scope by
+  -- 'Clash.Annotations.SynthesisAttributes.annotateReg'
+  deriving (Eq, Show, Generic, NFData, Binary)
+
+instance Ord TickInfo where
+  compare (SrcSpan s1) (SrcSpan s2) = leftmost_smallest s1 s2
+  compare (NameMod m1 t1) (NameMod m2 t2) =
+    compare m1 m2 `thenCompare` compare t1 t2
+  compare (Attributes t1 a1) (Attributes t2 a2) =
+    compare t1 t2 `thenCompare` acmpTerm a1 a2
+  compare t1 t2 = compare (getRank t1) (getRank t2)
+    where
+      getRank :: TickInfo -> Word
+      getRank SrcSpan{}     = 0
+      getRank NameMod{}     = 1
+      getRank DeDup         = 2
+      getRank NoDeDup       = 3
+      getRank Attributes {} = 4
 
 -- | Tag to indicate which instance/register name modifier was used
 data NameMod
@@ -119,12 +154,12 @@ data NameMod
   -- ^ @Clash.Magic.suffixNameP@
   | SetName
   -- ^ @Clash.Magic.setName@
-  deriving (Eq,Show,Generic,NFData,Hashable,Binary)
+  deriving (Eq,Ord,Show,Generic,NFData,Hashable,Binary)
 
 data IsMultiPrim
   = SingleResult
   | MultiResult
-  deriving (Show, Generic, NFData, Hashable, Binary)
+  deriving (Show, Generic, NFData, Eq, Hashable, Binary)
 
 data PrimInfo = PrimInfo
   { primName :: !Text
@@ -137,7 +172,13 @@ data PrimInfo = PrimInfo
   -- the variables it should assign its results to.
   --
   -- See: 'Clash.Normalize.Transformations.setupMultiResultPrim'
-  } deriving (Show,Generic,NFData,Hashable,Binary)
+  , primUnfolding :: !PrimUnfolding
+  } deriving (Show, Generic, NFData, Binary)
+
+data PrimUnfolding
+  = NoUnfolding
+  | Unfolding !Id
+  deriving (Show, Generic, NFData, Eq, Hashable, Binary)
 
 data MultiPrimInfo = MultiPrimInfo
   { mpi_primInfo :: PrimInfo
@@ -155,6 +196,10 @@ data WorkInfo
   | WorkAlways
   -- ^ Performs work regardless of whether the variables are constant or
   -- variable; these are things like clock or reset generators
+  | WorkIdentity Int [Int]
+  -- ^ A more restrictive version of 'WorkNever', where the value is the
+  -- argument at the given position if all arguments for the given list of
+  -- positions are also 'WorkIdentity'
   deriving (Eq,Show,Generic,NFData,Hashable,Binary)
 
 -- | Term reference
@@ -162,6 +207,12 @@ type TmName     = Name Term
 -- | Binding in a LetRec construct
 type LetBinding = (Id, Term)
 
+data Bind a
+  = NonRec Id a
+  | Rec [(Id, a)]
+  deriving (Eq, Show, Generic, NFData, Hashable, Binary, Functor)
+  -- Structural equivalence instead of alpha equivalance
+
 -- | Patterns in the LHS of a case-decomposition
 data Pat
   = DataPat !DataCon [TyVar] [Id]
@@ -171,7 +222,7 @@ data Pat
   -- ^ Literal pattern
   | DefaultPat
   -- ^ Default pattern
-  deriving (Eq,Ord,Show,Generic,NFData,Hashable,Binary)
+  deriving (Eq, Ord, Show, Generic, NFData, Binary)
 
 type Alt = (Pat,Term)
 
@@ -223,7 +274,7 @@ data CoreContext
   -- ^ Function position of a type application
   | LetBinding Id [Id]
   -- ^ RHS of a Let-binder with the sibling LHS'
-  | LetBody [Id]
+  | LetBody [LetBinding]
   -- ^ Body of a Let-binding with the bound LHS'
   | LamBody Id
   -- ^ Body of a lambda-term with the abstracted variable
@@ -237,7 +288,7 @@ data CoreContext
   -- ^ Body of a Cast
   | TickC TickInfo
   -- ^ Body of a Tick
-  deriving (Show, Generic, NFData, Hashable, Binary)
+  deriving (Show, Generic, NFData, Binary)
 
 -- | A list of @CoreContext@ describes the complete navigation path from the
 -- top-level to a specific sub-expression.
@@ -260,7 +311,7 @@ instance Eq CoreContext where
     -- NB: we do not see inside the argument here
     (TyAppC,          TyAppC)            -> True
     (LetBinding i is, LetBinding i' is') -> i == i' && is == is'
-    (LetBody is,      LetBody is')       -> is == is'
+    (LetBody is,      LetBody is')       -> map fst is == map fst is'
     (LamBody i,       LamBody i')        -> i == i'
     (TyLamBody tv,    TyLamBody tv')     -> tv == tv'
     (CaseAlt p,       CaseAlt p')        -> p == p'
@@ -283,6 +334,23 @@ stripTicks :: Term -> Term
 stripTicks (Tick _ e) = stripTicks e
 stripTicks e = e
 
+-- | Like 'stripTicks' but removes all ticks from subexpressions.
+stripAllTicks :: Term -> Term
+stripAllTicks = go
+ where
+  go (Lam i x) = Lam i (go x)
+  go (TyLam i x) = TyLam i (go x)
+  go (App f x) = App (go f) (go x)
+  go (TyApp f a) = TyApp (go f) a
+  go (Let bs x) = Let (goBinds bs) (go x)
+  go (Case x ty alts) = Case (go x) ty (fmap go <$> alts)
+  go (Cast x a b) = Cast (go x) a b
+  go (Tick _ x) = go x
+  go x = x
+
+  goBinds (NonRec i x) = NonRec i (go x)
+  goBinds (Rec ixs) = Rec (fmap go <$> ixs)
+
 -- | Split a (Type)Application in the applied term and it arguments
 collectArgs :: Term -> (Term, [Either Term Type])
 collectArgs = go []
@@ -350,7 +418,8 @@ walkTerm f = catMaybes . DList.toList . go
     TyLam _ t1 -> go t1
     App t1 t2 -> go t1 <> go t2
     TyApp t1 _ -> go t1
-    Letrec bndrs t1 -> go t1 <> mconcat (map (go . snd) bndrs)
+    Let (NonRec _ x) t1 -> go t1 <> go x
+    Let (Rec bndrs) t1 -> go t1 <> mconcat (map (go . snd) bndrs)
     Case t1 _ alts -> go t1 <> mconcat (map (go . snd) alts)
     Cast t1 _ _ -> go t1
     Tick _ t1 -> go t1
@@ -362,7 +431,8 @@ collectTermIds = concat . walkTerm (Just . go)
   go :: Term -> [Id]
   go (Var i) = [i]
   go (Lam i _) = [i]
-  go (Letrec bndrs _) = map fst bndrs
+  go (Let (NonRec i _) _) = [i]
+  go (Let (Rec bndrs) _) = fmap fst bndrs
   go (Case _ _ alts) = concatMap (pat . fst) alts
   go (Data _) = []
   go (Literal _) = []
@@ -378,12 +448,10 @@ collectTermIds = concat . walkTerm (Just . go)
   pat (LitPat _) = []
   pat DefaultPat = []
 
--- | Make variable reference out of term variable
-idToVar :: Id -> Term
-idToVar i@(Id {}) = Var i
-idToVar tv        = error $ $(curLoc) ++ "idToVar: tyVar: " ++ show tv
-
--- | Make a term variable out of a variable reference
+-- | Make a term variable out of a variable reference or ticked variable
+-- reference
 varToId :: Term -> Id
-varToId (Var i) = i
-varToId e       = error $ $(curLoc) ++ "varToId: not a var: " ++ show e
+varToId = \case
+  Var i    -> i
+  Tick _ e -> varToId e
+  e        -> error $ $(curLoc) ++ "varToId: not a var: " ++ show e
