@@ -1,8 +1,10 @@
 {-|
   Copyright  :  (C) 2012-2016, University of Twente,
-                    2017     , Myrtle Software Ltd
+                    2017     , Myrtle Software Ltd,
+                    2021-2022, QBayLogic B.V.
+                    2022     , Google Inc.
   License    :  BSD2 (see the file LICENSE)
-  Maintainer :  Christiaan Baaij <christiaan.baaij@gmail.com>
+  Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
   Parser definitions for BlackBox templates
 -}
@@ -16,7 +18,6 @@ where
 import           Control.Applicative          ((<|>))
 import           Data.Text.Lazy               (Text, pack, unpack)
 import qualified Data.Text.Lazy               as Text
-import           Text.Parser.Combinators
 import           Text.Trifecta                hiding (Err)
 import           Text.Trifecta.Delta
 
@@ -25,11 +26,8 @@ import           Clash.Netlist.BlackBox.Types
 
 -- | Parse a text as a BlackBoxTemplate, returns a list of errors in case
 -- parsing fails
--- runParse :: Text -> (BlackBoxTemplate, [Error LineColPos])
--- runParse = PCC.parse ((,) <$> pBlackBoxD <*> pEnd)
---          . createStr (LineColPos 0 0 0)
 runParse :: Text -> Result BlackBoxTemplate
-runParse = parseString pBlackBoxD (Directed "" 0 0 0 0) . unpack
+runParse = parseString (pBlackBoxD <* eof) (Directed "" 0 0 0 0) . unpack
 
 -- | Parse a BlackBoxTemplate (Declarations and Expressions)
 pBlackBoxD :: Parser BlackBoxTemplate
@@ -80,7 +78,7 @@ pTagE =  Result            <$  string "~RESULT"
      <|> Const             <$> (string "~CONST" *> brackets' natural')
      <|> Lit               <$> (string "~LIT" *> brackets' natural')
      <|> Name              <$> (string "~NAME" *> brackets' natural')
-     <|> ToVar             <$> try (string "~VAR" *> brackets' pSigD) <*> brackets' natural'
+     <|> ToVar             <$> try (string "~VAR" *> brackets' pSigDorEmpty) <*> brackets' natural'
      <|> (Sym Text.empty)  <$> (string "~SYM" *> brackets' natural')
      <|> Typ Nothing       <$  string "~TYPO"
      <|> (Typ . Just)      <$> try (string "~TYP" *> brackets' natural')
@@ -110,11 +108,14 @@ pTagE =  Result            <$  string "~RESULT"
      <|> Sel               <$> (string "~SEL" *> brackets' pTagE) <*> brackets' natural'
      <|> IsLit             <$> (string "~ISLIT" *> brackets' natural')
      <|> IsVar             <$> (string "~ISVAR" *> brackets' natural')
+     <|> IsScalar          <$> (string "~ISSCALAR" *> brackets' natural')
      <|> IsActiveHigh      <$> (string "~ISACTIVEHIGH" *> brackets' natural')
      <|> IsActiveEnable    <$> (string "~ISACTIVEENABLE" *> brackets' natural')
      <|> IsUndefined       <$> (string "~ISUNDEFINED" *> brackets' natural')
      <|> StrCmp            <$> (string "~STRCMP" *> brackets' pSigD) <*> brackets' natural'
-     <|> OutputWireReg     <$> (string "~OUTPUTWIREREG" *> brackets' natural')
+     -- Parse ~OUTPUTWIREREG for backwards compatibility
+     <|> OutputUsage       <$> (string "~OUTPUTWIREREG" *> brackets' natural')
+     <|> OutputUsage       <$> (string "~OUTPUTUSAGE" *> brackets' natural')
      <|> GenSym            <$> (string "~GENSYM" *> brackets' pSigD) <*> brackets' natural'
      <|> Template          <$> (string "~TEMPLATE" *> brackets' pSigD) <*> brackets' pSigD
      <|> Repeat            <$> (string "~REPEAT" *> brackets' pSigD) <*> brackets' pSigD
@@ -129,6 +130,7 @@ pTagE =  Result            <$  string "~RESULT"
      <|> IsSync            <$> (string "~ISSYNC" *> brackets' natural')
      <|> IsInitDefined     <$> (string "~ISINITDEFINED" *> brackets' natural')
      <|> CtxName           <$  string "~CTXNAME"
+     <|> LongestPeriod     <$  string "~LONGESTPERIOD"
 
 natural' :: TokenParsing m => m Int
 natural' = fmap fromInteger natural
@@ -148,7 +150,10 @@ pElemE = pTagE
 
 -- | Parse SigD
 pSigD :: Parser [Element]
-pSigD = some (pTagE <|> (Text (pack "[") <$ (pack <$> string "[\\"))
-                    <|> (Text (pack "]") <$ (pack <$> string "\\]"))
+pSigD = some (pTagE <|> (EscapedSymbol SquareBracketOpen <$ string "[\\")
+                    <|> (EscapedSymbol SquareBracketClose <$ string "\\]")
                     <|> (Text <$> (pack <$> some (satisfyRange '\000' '\90')))
                     <|> (Text <$> (pack <$> some (satisfyRange '\94' '\125'))))
+
+pSigDorEmpty :: Parser [Element]
+pSigDorEmpty = pSigD <|> mempty
