@@ -2,7 +2,8 @@
   Copyright  :  (C) 2012-2016, University of Twente,
                     2017     , Myrtle Software Ltd,
                     2017-2018, Google Inc.
-                    2020     , QBayLogic
+                    2020-2023, QBayLogic B.V.
+                    2022-2023, Google Inc.
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -21,12 +22,6 @@
 {-# LANGUAGE PatternSynonyms #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TemplateHaskell #-}
-{-# LANGUAGE TypeSynonymInstances #-}
-
--- since GHC 8.6 we can haddock individual constructor fields \o/
-#if __GLASGOW_HASKELL__ >= 806
-#define FIELD ^
-#endif
 
 module Clash.Netlist.Types
   ( Declaration (..,NetDecl)
@@ -44,6 +39,8 @@ import Control.Monad.Reader                 (ReaderT, MonadReader)
 import qualified Control.Monad.State        as Lazy (State)
 import qualified Control.Monad.State.Strict as Strict
   (State, MonadIO, MonadState, StateT)
+import Data.Aeson                           (FromJSON(..))
+import qualified Data.Aeson as Aeson
 import Data.Bits                            (testBit)
 import Data.Binary                          (Binary(..))
 import Data.Function                        (on)
@@ -53,13 +50,15 @@ import Data.HashSet                         (HashSet)
 import qualified Data.List                  as List
 import Data.IntMap                          (IntMap, empty)
 import Data.Map.Ordered                     (OMap)
+import Data.Map                             (Map)
+import qualified Data.Map as Map
 import Data.Maybe                           (mapMaybe)
+import Data.Monoid                          (Ap(..))
 import qualified Data.Set                   as Set
 import Data.Text                            (Text)
 
 import Data.Typeable                        (Typeable)
 import Data.Text.Prettyprint.Doc.Extra      (Doc)
-import Data.Semigroup.Monad                 (Mon(..))
 import GHC.Generics                         (Generic)
 import GHC.Stack
 import Language.Haskell.TH.Syntax           (Lift)
@@ -70,26 +69,28 @@ import GHC.Types.SrcLoc                     (SrcSpan)
 import SrcLoc                               (SrcSpan)
 #endif
 
+import Clash.Annotations.SynthesisAttributes(Attr)
 import Clash.Annotations.BitRepresentation  (FieldAnn)
 import Clash.Annotations.Primitive          (HDL(..))
 import Clash.Annotations.TopEntity          (TopEntity)
-import Clash.Backend                        (Backend)
+import Clash.Backend                        (Backend, HasUsageMap (..))
+import Clash.Core.HasType
+import Clash.Core.PartialEval               (Evaluator)
 import Clash.Core.Type                      (Type)
-import Clash.Core.Var                       (Attr', Id, varType)
+import Clash.Core.Var                       (Id)
 import Clash.Core.TyCon                     (TyConMap)
 import Clash.Core.VarEnv                    (VarEnv)
-import Clash.Driver.Types                   (BindingMap, ClashOpts)
+import Clash.Driver.Types                   (BindingMap, ClashEnv(..), ClashOpts(..))
 import Clash.Netlist.BlackBox.Types         (BlackBoxTemplate)
 import Clash.Primitives.Types               (CompiledPrimMap)
 import Clash.Signal.Internal
   (ResetPolarity, ActiveEdge, ResetKind, InitBehavior)
 import Clash.Unique                         (Unique)
-import Clash.Util                           (makeLenses)
 
 import Clash.Annotations.BitRepresentation.Internal
   (CustomReprs, DataRepr', ConstrRepr')
 
-import {-# SOURCE #-} qualified Clash.Netlist.Id as Id
+import {-# SOURCE #-} qualified Clash.Netlist.Id as Id (toText)
 
 -- | Structure describing a top entity: it's id and its port annotations.
 data TopEntityT = TopEntityT
@@ -99,7 +100,7 @@ data TopEntityT = TopEntityT
   -- ^ (Maybe) a topentity annotation
   , topIsTestBench :: Bool
   -- ^ Whether this entity is a test bench
-  } deriving (Generic, Show)
+  } deriving (Generic, Show, Eq)
 
 -- | Same as "TopEntity", but with all port names that end up in HDL specified
 data ExpandedTopEntity a = ExpandedTopEntity
@@ -118,11 +119,11 @@ data ExpandedPortName a
   -- | Same as "PortProduct", but fully expanded
   | ExpandedPortProduct
       Text
-      -- FIELD Name hint. Can be used to create intermediate signal names.
+      -- ^ Name hint. Can be used to create intermediate signal names.
       HWType
-      -- FIELD Type of product
+      -- ^ Type of product
       [ExpandedPortName a]
-      -- FIELD Product fields
+      -- ^ Product fields
   deriving (Show, Functor, Foldable, Traversable)
 
 -- | Monad that caches generated components (StateT) and remembers hidden inputs
@@ -132,7 +133,7 @@ newtype NetlistMonad a =
   deriving newtype (Functor, Monad, Applicative, MonadReader NetlistEnv,
                     Strict.MonadState NetlistState, Strict.MonadIO, MonadFail)
 
-type HWMap = HashMap Type (Either String FilteredHWType)
+type HWMap = Map Type (Either String FilteredHWType)
 
 -- | See 'is_freshCache'
 type FreshCache = HashMap Text (IntMap Word)
@@ -207,16 +208,16 @@ data IdentifierSet
 -- identifiers internally. Any Identifier should be trivially printable to any
 -- HDL.
 --
--- __NB__: use the functions in Clash.Netlist.Id. Don't use these constructors
+-- __NB__: use the functions in "Clash.Netlist.Id". Don't use these constructors
 -- directly.
 data Identifier
   -- | Unparsed identifier. Used for things such as port names, which should
   -- appear in the HDL exactly as the user specified.
   = RawIdentifier
       !Text
-      -- FIELD An identifier exactly as given by the user
+      -- ^ An identifier exactly as given by the user
       (Maybe Identifier)
-      -- FIELD Parsed version of raw identifier. Will not be populated if this
+      -- ^ Parsed version of raw identifier. Will not be populated if this
       -- identifier was created with an unsafe function.
       !CallStack
       -- ^ Stores where this identifier was generated. Tracking is only enabled
@@ -241,7 +242,7 @@ data Identifier
     -- name of 'foo' and an extension of [6, 5] would render as 'foo_5_6'. Note
     -- that extensions are stored in reverse order for easier manipulation.
     , i_idType :: !IdentifierType
-    -- ^ See "IdentifierType".
+    -- ^ See 'IdentifierType'.
     , i_hdl :: !HDL
     -- ^ HDL this identifier is generated for.
     , i_provenance :: !CallStack
@@ -275,18 +276,25 @@ instance Ord Identifier where
 -- | Environment of the NetlistMonad
 data NetlistEnv
   = NetlistEnv
-  { _prefixName  :: Text
+  { _clashEnv :: ClashEnv
+  , _prefixName  :: Text
   -- ^ Prefix for instance/register names
   , _suffixName :: Text
   -- ^ Postfix for instance/register names
   , _setName :: Maybe Text
   -- ^ (Maybe) user given instance/register name
+  , _localAttrs :: [Attr Text]
+  -- ^ Synthesis attributes brough into scope by
+  -- 'Clash.Annotations.SynthesisAttributes.annotateReg'
+  , _peEvaluator :: Evaluator
+  -- ^ Evaluator to evaluate a term to Normal Form
   }
 
 data ComponentMeta = ComponentMeta
   { cmWereVoids :: [Bool]
   , cmLoc :: SrcSpan
   , cmScope :: IdentifierSet
+  , cmUsage :: UsageMap
   } deriving (Generic, Show, NFData)
 
 type ComponentMap = OMap Unique (ComponentMeta, Component)
@@ -299,15 +307,10 @@ data NetlistState
   , _components     :: ComponentMap
   -- ^ Cached components. Is an insertion ordered map to preserve a topologically
   -- sorted component list for the manifest file.
-  , _primitives     :: CompiledPrimMap
-  -- ^ Primitive Definitions
   , _typeTranslator :: CustomReprs -> TyConMap -> Type
                     -> Strict.State HWMap (Maybe (Either String FilteredHWType))
   -- ^ Hardcoded Type -> HWType translator
-  , _tcCache        :: TyConMap
-  -- ^ TyCon cache
   , _curCompNm      :: !(Identifier,SrcSpan)
-  , _intWidth       :: Int
   , _seenIds        :: IdentifierSet
   -- ^ All names currently in scope.
   , _seenComps      :: IdentifierSet
@@ -340,9 +343,6 @@ data NetlistState
   , _hdlDir         :: FilePath
   , _curBBlvl       :: Int
   -- ^ The current scoping level assigned to black box contexts
-  , _customReprs    :: CustomReprs
-  , _clashOpts      :: ClashOpts
-  -- ^ Settings Clash was called with
   , _isTestBench    :: Bool
   -- ^ Whether we're compiling a testbench (suppresses some warnings)
   , _backEndITE :: Bool
@@ -350,6 +350,10 @@ data NetlistState
   , _backend :: SomeBackend
   -- ^ The current HDL backend
   , _htyCache :: HWMap
+  , _usages :: UsageMap
+  -- ^ The current way signals are assigned in netlist. This is used to
+  -- determine how signals are rendered in HDL (i.e. wire/reg in Verilog, or
+  -- signal/variable in VHDL).
   }
 
 data ComponentPrefix
@@ -365,24 +369,36 @@ data SomeBackend where
   SomeBackend :: Backend backend => backend -> SomeBackend
 
 type Comment = Text
+type Directive = Text
+
+data CommentOrDirective
+  = Comment Comment
+  | Directive Directive
+  deriving Show
 
 -- | Component: base unit of a Netlist
 data Component
   = Component
   { componentName :: !Identifier -- ^ Name of the component
   , inputs        :: [(Identifier,HWType)] -- ^ Input ports
-  , outputs       :: [(WireOrReg,(Identifier,HWType),Maybe Expr)] -- ^ Output ports
+  , outputs       :: [(Usage,(Identifier,HWType),Maybe Expr)] -- ^ Output ports
   , declarations  :: [Declaration] -- ^ Internal declarations
   }
-  deriving Show
+  deriving (Show, Generic, NFData)
 
-instance NFData Component where
-  rnf c = case c of
-    Component nm inps outps decls -> rnf nm    `seq` rnf inps `seq`
-                                     rnf outps `seq` rnf decls
+-- | Check if an input port is really an inout port.
+--
+isBiDirectional :: (Identifier, HWType) -> Bool
+isBiDirectional = go . snd
+  where
+    go BiDirectional{} = True
+    go (Annotated _ hwty) = go hwty
+    go _ = False
 
 -- | Find the name and domain name of each clock argument of a component.
 --
+-- This will not consider @ClockN@ to be a clock argument, which means only the
+-- positive phase of a differential pair will be added to @sdcClock@.
 findClocks :: Component -> [(Text, Text)]
 findClocks (Component _ is _ _) =
   mapMaybe isClock is
@@ -429,6 +445,8 @@ data HWType
   -- ^ Unsigned integer of a specified size
   | Vector !Size !HWType
   -- ^ Vector type
+  | MemBlob !Size !Size
+  -- ^ MemBlob type
   | RTree !Size !HWType
   -- ^ RTree type
   | Sum !Text [Text]
@@ -440,6 +458,8 @@ data HWType
   -- ^ Sum-of-Product type: Name and Constructor names + field types
   | Clock !DomainName
   -- ^ Clock type corresponding to domain /DomainName/
+  | ClockN !DomainName
+  -- ^ ClockN type corresponding to domain /DomainName/
   | Reset !DomainName
   -- ^ Reset type corresponding to domain /DomainName/
   | Enable !DomainName
@@ -455,7 +475,7 @@ data HWType
   | CustomProduct !Text !DataRepr' !Size (Maybe [Text]) [(FieldAnn, HWType)]
   -- ^ Same as Product, but with a user specified bit representation. For more
   -- info, see: Clash.Annotations.BitRepresentations.
-  | Annotated [Attr'] !HWType
+  | Annotated [Attr Text] !HWType
   -- ^ Annotated with HDL attributes
   | KnownDomain !DomainName !Integer !ActiveEdge !ResetKind !InitBehavior !ResetPolarity
   -- ^ Domain name, period, active edge, reset kind, initial value behavior
@@ -463,9 +483,17 @@ data HWType
   -- ^ File type for simulation-level I/O
   deriving (Eq, Ord, Show, Generic, NFData, Hashable)
 
+-- | Smart constructor for 'Annotated'. Wraps the given type in an 'Annotated'
+-- if the attribute list is non-empty. If it is empty, it will return the given
+-- 'HWType' unchanged.
+annotated :: [Attr Text] -> HWType -> HWType
+annotated [] t = t
+annotated attrs t = Annotated attrs t
+
 hwTypeDomain :: HWType -> Maybe DomainName
 hwTypeDomain = \case
   Clock dom -> Just dom
+  ClockN dom -> Just dom
   Reset dom -> Just dom
   Enable dom -> Just dom
   KnownDomain dom _ _ _ _ _ -> Just dom
@@ -473,7 +501,7 @@ hwTypeDomain = \case
 
 -- | Extract hardware attributes from Annotated. Returns an empty list if
 -- non-Annotated given or if Annotated has an empty list of attributes.
-hwTypeAttrs :: HWType -> [Attr']
+hwTypeAttrs :: HWType -> [Attr Text]
 hwTypeAttrs (Annotated attrs _type) = attrs
 hwTypeAttrs _                       = []
 
@@ -505,81 +533,195 @@ data PortMap
 data Declaration
   -- | Signal assignment
   = Assignment
-      !Identifier -- FIELD Signal to assign
-      !Expr       -- FIELD Assigned expression
+      !Identifier -- ^ Signal to assign
+      !Usage      -- ^ How the signal is assigned
+      !Expr       -- ^ Assigned expression
 
   -- | Conditional signal assignment:
   | CondAssignment
-      !Identifier            -- FIELD Signal to assign
-      !HWType                -- FIELD Type of the result/alternatives
-      !Expr                  -- FIELD Scrutinized expression
-      !HWType                -- FIELD Type of the scrutinee
-      [(Maybe Literal,Expr)] -- FIELD List of: (Maybe expression scrutinized expression is compared with,RHS of alternative)
+      !Identifier            -- ^ Signal to assign
+      !HWType                -- ^ Type of the result/alternatives
+      !Expr                  -- ^ Scrutinized expression
+      !HWType                -- ^ Type of the scrutinee
+      [(Maybe Literal,Expr)] -- ^ List of: (Maybe expression scrutinized expression is compared with,RHS of alternative)
 
   -- | Instantiation of another component:
   | InstDecl
-      EntityOrComponent                  -- FIELD Whether it's an entity or a component
-      (Maybe Text)                       -- FIELD Library instance is defined in
-      [Attr']                            -- FIELD Attributes to add to the generated code
-      !Identifier                        -- FIELD The component's (or entity's) name
-      !Identifier                        -- FIELD Instance label
-      [(Expr,HWType,Expr)]               -- FIELD List of parameters for this component (param name, param type, param value)
+      EntityOrComponent                  -- ^ Whether it's an entity or a component
+      (Maybe Text)                       -- ^ Library instance is defined in
+      [Attr Text]                        -- ^ Attributes to add to the generated code
+      !Identifier                        -- ^ The component's (or entity's) name
+      !Identifier                        -- ^ Instance label
+      [(Expr,HWType,Expr)]               -- ^ List of parameters for this component (param name, param type, param value)
       PortMap
 
   -- | Instantiation of blackbox declaration
   | BlackBoxD
-      !Text                    -- FIELD Primitive name
-      [BlackBoxTemplate]       -- FIELD VHDL only: add @library@ declarations
-      [BlackBoxTemplate]       -- FIELD VHDL only: add @use@ declarations
-      [((Text,Text),BlackBox)] -- FIELD Intel Quartus only: create a @.qsys@ file from given template
-      !BlackBox                -- FIELD Template tokens
-      BlackBoxContext          -- FIELD Context in which tokens should be rendered
+      !Text                    -- ^ Primitive name
+      [BlackBoxTemplate]       -- ^ VHDL only: add @library@ declarations
+      [BlackBoxTemplate]       -- ^ VHDL only: add @use@ declarations
+      [((Text,Text),BlackBox)] -- ^ Intel Quartus only: create a @.qsys@ file from given template
+      !BlackBox                -- ^ Template tokens
+      BlackBoxContext          -- ^ Context in which tokens should be rendered
+
+  -- | @component@ declaration (VHDL).
+  --
+  -- See [this tutorial](https://www.ics.uci.edu/~jmoorkan/vhdlref/compdec.html);
+  -- refer to §4.5 of IEEE 1076-1993
+  | CompDecl
+      !Text
+      [(Text, PortDirection, HWType)]
 
   -- | Signal declaration
   | NetDecl'
-      (Maybe Comment)                -- FIELD Note; will be inserted as a comment in target hdl
-      WireOrReg                      -- FIELD Wire or register
-      !Identifier                    -- FIELD Name of signal
-      (Either IdentifierText HWType) -- FIELD Pointer to type of signal or type of signal
-      (Maybe Expr)                   -- FIELD Initial value
+      (Maybe Comment)                -- ^ Note; will be inserted as a comment in target hdl
+      !Identifier                    -- ^ Name of signal
+      HWType                         -- ^ Type of signal
+      (Maybe Expr)                   -- ^ Initial value
       -- ^ Signal declaration
-  | TickDecl Comment
-  -- ^ HDL tick corresponding to a Core tick
+
+  -- | HDL tick corresponding to a Core tick
+  | TickDecl CommentOrDirective
+
   -- | Sequential statement
   | Seq [Seq]
+
+  -- | Compilation conditional on some preprocessor symbol, note that
+  -- declarations here are ignored for VHDL. See here for a discussion
+  -- https://github.com/clash-lang/clash-compiler/pull/1798#discussion_r648571862
+  | ConditionalDecl
+      !Text -- ^ condition text, for example @FORMAL@
+      [Declaration]
   deriving Show
 
 -- | Sequential statements
 data Seq
   -- | Clocked sequential statements
   = AlwaysClocked
-      ActiveEdge -- FIELD Edge of the clock the statement should be executed
-      Expr       -- FIELD Clock expression
-      [Seq]      -- FIELD Statements to be executed on the active clock edge
+      ActiveEdge -- ^ Edge of the clock the statement should be executed
+      Expr       -- ^ Clock expression
+      [Seq]      -- ^ Statements to be executed on the active clock edge
   -- | Statements running at simulator start
   | Initial
-      [Seq] -- FIELD Statements to run at simulator start
+      [Seq] -- ^ Statements to run at simulator start
   -- | Statements to run always
   | AlwaysComb
-      [Seq] -- FIELD Statements to run always
+      [Seq] -- ^ Statements to run always
   -- | Declaration in sequential form
   | SeqDecl
-      Declaration -- FIELD The declaration
+      Declaration -- ^ The declaration
   -- | Branching statement
   | Branch
-      !Expr                    -- FIELD Scrutinized expresson
-      !HWType                  -- FIELD Type of the scrutinized expression
-      [(Maybe Literal,[Seq])]  -- FIELD List of: (Maybe match, RHS of Alternative)
+      !Expr                    -- ^ Scrutinized expresson
+      !HWType                  -- ^ Type of the scrutinized expression
+      [(Maybe Literal,[Seq])]  -- ^ List of: (Maybe match, RHS of Alternative)
   deriving Show
 
+-- | Procedural assignment in HDL can be blocking or non-blocking. This
+-- determines when the assignment takes place in simulation. The name refers to
+-- whether evaluation of the remaining statements in a process is blocked
+-- until the assignment is performed or not.
+--
+-- See Also:
+--
+-- IEEE 1364-2001, sections 9.2.1 and 9.2.2
+-- IEEE 1076-1993, sections 8.4 and 8.5
+--
+data Blocking
+  = NonBlocking
+  -- ^ A non-blocking assignment means the new value is not observed until the
+  -- next time step in simulation. Using the signal later in the process will
+  -- continue to return the old value.
+  | Blocking
+  -- ^ A blocking assignment means the new value is observed immediately. Using
+  -- the signal later in the process will return the new value.
+  deriving (Binary, Eq, Generic, Hashable, NFData, Show)
+
+-- NOTE [`Semigroup` instances for `Blocking` and `Usage`]
+instance Semigroup Blocking where
+  NonBlocking <> y = y
+  Blocking    <> _ = Blocking
+
+-- | The usage of a signal refers to how the signal is written to in netlist.
+-- This is used to determine if the signal should be a @wire@ or @reg@ in
+-- (System)Verilog, or a @signal@ or @variable@ in VHDL.
+--
+data Usage
+  = Cont
+  -- ^ Continuous assignment, which occurs in a concurrent context.
+  | Proc Blocking
+  -- ^ Procedural assignment, which occurs in a sequential context.
+  deriving (Binary, Eq, Generic, Hashable, NFData, Show)
+
+-- NOTE [`Semigroup` instances for `Blocking` and `Usage`]
+instance Semigroup Usage where
+  Cont    <> y      = y
+  Proc x  <> Proc y = Proc (x <> y)
+  Proc x  <> _      = Proc x
+
+{-
+NOTE [`Semigroup` instances for `Blocking` and `Usage`]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Usages (and Blocking) are combined by taking the most restrictive usage, where
+most restrictive means "has the most influence over the choice of declaration".
+Clash produces three types of assignment:
+
+  * continuous
+  * prodecural non-blocking
+  * prodecural blocking
+
+Both VHDl and (System)Verilog have a type of declaration which only admits one
+type of assignment. This is the most restrictive for that HDL. However, since
+that would involve knowing the HDL type in these Semigroup instances, the
+most restrictive here is based on ordering where the most restrictive for each
+HDL is an extreme value (max for VHDL, min for Verilog). i.e.
+
+          |-------------------------------------|
+          | Continuous | NonBlocking | Blocking |
+|---------|-------------------------------------|
+|    VHDL |         signal           | variable |
+|---------|-------------------------------------|
+| Verilog |   wire     |          reg           |
+|---------|-------------------------------------|
+-}
+
+instance FromJSON Usage where
+  parseJSON = Aeson.withText "Usage" $ \case
+    "Continuous"  -> pure Cont
+    "NonBlocking" -> pure (Proc NonBlocking)
+    "Blocking"    -> pure (Proc Blocking)
+    str           -> fail $ mconcat
+      [ "Could not parse usage: "
+      , show str
+      , "\nRecognized values are 'Continuous', 'NonBlocking' and 'Blocking'"
+      ]
+
+-- See NOTE [`Text` key for `UsageMap`]
+type UsageMap = Map Text Usage
+
+lookupUsage :: Identifier -> UsageMap -> Maybe Usage
+lookupUsage i = Map.lookup (Id.toText i)
+
+{-
+NOTE [`Text` key for `UsageMap`]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+We would like to use netlist identifiers as the key for the usage map, since
+conceptually it is a map from an identifier to how it is used in assignments.
+However, in practice we commonly end up with the same textual identifier
+appearing in different ways in the netlist.
+
+The most obvious example of this are identifiers that appear as both
+`UniqueIdentifier` and `RawIdentifier`. If we track the usage on the raw
+identifier, but the `NetDecl` uses the `UniqueIdentifier` then the wrong
+declaration may be used in the rendered HDL.
+
+Attempting to fix this by not generating the same textual identifier in
+different ways proved difficult, so for now the key type is `Text` instead.
+-}
+
 data EntityOrComponent = Entity | Comp | Empty
   deriving Show
 
-data WireOrReg = Wire | Reg
-  deriving (Show,Generic)
-
-instance NFData WireOrReg
-
 pattern NetDecl
   :: Maybe Comment
   -- ^ Note; will be inserted as a comment in target hdl
@@ -588,9 +730,9 @@ pattern NetDecl
   -> HWType
   -- ^ Type of signal
   -> Declaration
-pattern NetDecl note d ty <- NetDecl' note Wire d (Right ty) _
+pattern NetDecl note d ty <- NetDecl' note d ty _
   where
-    NetDecl note d ty = NetDecl' note Wire d (Right ty) Nothing
+    NetDecl note d ty = NetDecl' note d ty Nothing
 
 data PortDirection = In | Out
   deriving (Eq,Ord,Show,Generic,NFData,Hashable)
@@ -600,11 +742,18 @@ instance NFData Declaration where
 
 -- | Expression Modifier
 data Modifier
-  = Indexed (HWType,Int,Int) -- ^ Index the expression: (Type of expression,DataCon tag,Field Tag)
-  | DC (HWType,Int)          -- ^ See expression in a DataCon context: (Type of the expression, DataCon tag)
-  | VecAppend                -- ^ See the expression in the context of a Vector append operation
-  | RTreeAppend              -- ^ See the expression in the context of a Tree append operation
-  | Sliced (HWType,Int,Int)  -- ^ Slice the identifier of the given type from start to end
+  = Indexed (HWType, Int, Int)
+  -- ^ Index the expression: (Type of expression, DataCon tag, Field Tag). Note
+  -- that the type of the expression is the type we are slicing from, not the type
+  -- returned by the index operation.
+  | DC (HWType, Int)
+  -- ^ See expression in a DataCon context: (Type of the expression, DataCon tag)
+  | VecAppend
+  -- ^ See the expression in the context of a Vector append operation
+  | RTreeAppend
+  -- ^ See the expression in the context of a Tree append operation
+  | Sliced (HWType, Int, Int)
+  -- ^ Slice the identifier of the given type from start to end
   | Nested Modifier Modifier
   deriving Show
 
@@ -617,25 +766,25 @@ data Expr
 
   -- | Instantiation of a BlackBox expression
   | BlackBoxE
-      !Text                    -- FIELD Primitive name
-      [BlackBoxTemplate]       -- FIELD VHDL only: add @library@ declarations
-      [BlackBoxTemplate]       -- FIELD VHDL only: add @use@ declarations:
-      [((Text,Text),BlackBox)] -- FIELD Intel/Quartus only: create a @.qsys@ file from given template.
-      !BlackBox                -- FIELD Template tokens
-      !BlackBoxContext         -- FIELD Context in which tokens should be rendered
-      !Bool                    -- FIELD Wrap in parentheses?
+      !Text                    -- ^ Primitive name
+      [BlackBoxTemplate]       -- ^ VHDL only: add @library@ declarations
+      [BlackBoxTemplate]       -- ^ VHDL only: add @use@ declarations:
+      [((Text,Text),BlackBox)] -- ^ Intel/Quartus only: create a @.qsys@ file from given template.
+      !BlackBox                -- ^ Template tokens
+      !BlackBoxContext         -- ^ Context in which tokens should be rendered
+      !Bool                    -- ^ Wrap in parentheses?
 
   -- | Convert some type to a BitVector.
   | ToBv
-      (Maybe Identifier) -- FIELD Type prefix
-      HWType             -- FIELD Type to convert _from_
-      Expr               -- FIELD Expression to convert to BitVector
+      (Maybe Identifier) -- ^ Type prefix
+      HWType             -- ^ Type to convert _from_
+      Expr               -- ^ Expression to convert to BitVector
 
   -- | Convert BitVector to some type.
   | FromBv
-      (Maybe Identifier) -- FIELD Type prefix
-      HWType             -- FIELD Type to convert _to_
-      Expr               -- FIELD BitVector to convert
+      (Maybe Identifier) -- ^ Type prefix
+      HWType             -- ^ Type to convert _to_
+      Expr               -- ^ BitVector to convert
 
   | IfThenElse Expr Expr Expr
   -- | Do nothing
@@ -645,6 +794,24 @@ data Expr
 instance NFData Expr where
   rnf x = x `seq` ()
 
+isConstExpr :: Expr -> Bool
+isConstExpr = \case
+  Literal{} -> True
+  DataCon _ _ es -> all isConstExpr es
+  Identifier{} -> False
+  DataTag{} -> False
+  BlackBoxE nm _ _ _ _ ctx _
+    -- When using SimIO, `reg` creates (in Haskell) the mutable reference to
+    -- some value. The blackbox for this however is simply `~ARG[0]`, so if
+    -- the argument given is constant, the rendered HDL will also be constant.
+    | nm == "Clash.Explicit.SimIO.reg" ->
+        all (\(e, _, _) -> isConstExpr e) (bbInputs ctx)
+    | otherwise -> False
+  ToBv _ _ e -> isConstExpr e
+  FromBv _ _ e -> isConstExpr e
+  IfThenElse{} -> False
+  Noop -> False
+
 -- | Literals used in an expression
 data Literal
   = NumLit    !Integer          -- ^ Number literal
@@ -683,7 +850,7 @@ data BlackBoxContext
   , bbInputs :: [(Expr,HWType,Bool)]
   -- ^ Argument names, types, and whether it is a literal
   , bbFunctions :: IntMap [(Either BlackBox (Identifier,[Declaration])
-                          ,WireOrReg
+                          ,Usage
                           ,[BlackBoxTemplate]
                           ,[BlackBoxTemplate]
                           ,[((Text,Text),BlackBox)]
@@ -718,18 +885,20 @@ data BlackBox
 data TemplateFunction where
   TemplateFunction
     :: [Int]
-    -- FIELD Used arguments
+    -- ^ Used arguments
     -> (BlackBoxContext -> Bool)
-    -- FIELD Validation function. Should return 'False' if function can't render
+    -- ^ Validation function. Should return 'False' if function can't render
     -- given a certain context.
     -> (forall s . Backend s => BlackBoxContext -> Lazy.State s Doc)
-    -- FIELD Render function
+    -- ^ Render function
     -> TemplateFunction
 
 instance Show BlackBox where
-  show (BBTemplate t)  = "BBTemplate " <> show t
-  show (BBFunction nm hsh _) =
-    "<TemplateFunction(nm=" ++ show nm ++ ", hash=" ++ show hsh ++ ")>"
+  showsPrec d (BBTemplate t) =
+    showParen (d > 10) $ ("BBTemplate " ++) . showsPrec 11 t
+  showsPrec _ (BBFunction nm hsh _) =
+    ("<TemplateFunction(nm=" ++) . shows nm . (", hash=" ++) . shows hsh .
+    (")>" ++)
 
 instance NFData TemplateFunction where
   rnf (TemplateFunction is f _) = rnf is `seq` f `seq` ()
@@ -750,20 +919,7 @@ data NetlistId
   | MultiId [Id]
   -- ^ A split identifier (into several sub-identifiers), needed to assign
   -- expressions of types that have to be split apart (e.g. tuples of Files)
-  deriving Show
-
--- | Eliminator for 'NetlistId'
-netlistId
-  :: (Identifier -> r)
-  -- ^ Eliminator for Identifiers generated in the NetlistMonad
-  -> (Id -> r)
-  -- ^ Eliminator for original Core Identifiers
-  -> NetlistId
-  -> [r]
-netlistId f g = \case
-  NetlistId i _ -> [f i]
-  CoreId i -> [g i]
-  MultiId is -> map g is
+  deriving (Eq, Show)
 
 -- | Eliminator for 'NetlistId', fails on 'MultiId'
 netlistId1
@@ -786,8 +942,8 @@ netlistTypes
   -> [Type]
 netlistTypes = \case
   NetlistId _ t -> [t]
-  CoreId i -> [varType i]
-  MultiId is -> map varType is
+  CoreId i -> [coreTypeOf i]
+  MultiId is -> map coreTypeOf is
 
 -- | Return the type of a 'NetlistId', fails on 'MultiId'
 netlistTypes1
@@ -796,7 +952,7 @@ netlistTypes1
   -> Type
 netlistTypes1 = \case
   NetlistId _ t -> t
-  CoreId i -> varType i
+  CoreId i -> coreTypeOf i
   m -> error ("netlistTypes1 MultiId: " ++ show m)
 
 -- | Type of declaration, concurrent or sequential
@@ -804,6 +960,11 @@ data DeclarationType
   = Concurrent
   | Sequential
 
+-- | Default usage for a type of declaration (concurrent or sequential)
+declTypeUsage :: DeclarationType -> Usage
+declTypeUsage Concurrent = Cont
+declTypeUsage Sequential = Proc Blocking
+
 emptyBBContext :: Text -> BlackBoxContext
 emptyBBContext name
   = Context
@@ -819,8 +980,23 @@ emptyBBContext name
   , bbCtxName     = Nothing
   }
 
-makeLenses ''NetlistEnv
-makeLenses ''NetlistState
+Lens.makeLenses ''NetlistEnv
+Lens.makeLenses ''NetlistState
+
+intWidth :: Lens.Getter NetlistEnv Int
+intWidth = clashEnv . Lens.to (opt_intWidth . envOpts)
+
+customReprs :: Lens.Getter NetlistEnv CustomReprs
+customReprs = clashEnv . Lens.to envCustomReprs
+
+tcCache :: Lens.Getter NetlistEnv TyConMap
+tcCache = clashEnv . Lens.to envTyConMap
+
+primitives :: Lens.Getter NetlistEnv CompiledPrimMap
+primitives = clashEnv . Lens.to envPrimitives
+
+clashOpts :: Lens.Getter NetlistEnv ClashOpts
+clashOpts = clashEnv . Lens.to envOpts
 
 -- | Structures that hold an 'IdentifierSet'
 class HasIdentifierSet s where
@@ -829,6 +1005,12 @@ class HasIdentifierSet s where
 instance HasIdentifierSet IdentifierSet where
   identifierSet = ($)
 
+instance HasUsageMap NetlistState where
+  usageMap = usages
+
+instance HasIdentifierSet s => HasIdentifierSet (s, a) where
+  identifierSet = Lens._1 . identifierSet
+
 -- | An "IdentifierSetMonad" supports unique name generation for Clash Netlist
 class Monad m => IdentifierSetMonad m where
   identifierSetM :: (IdentifierSet -> IdentifierSet) -> m IdentifierSet
@@ -855,5 +1037,5 @@ instance HasIdentifierSet s => IdentifierSetMonad (Lazy.State s) where
     Lens.use identifierSet
   {-# INLINE identifierSetM #-}
 
-instance IdentifierSetMonad m => IdentifierSetMonad (Mon m) where
-  identifierSetM = Mon . identifierSetM
+instance IdentifierSetMonad m => IdentifierSetMonad (Ap m) where
+  identifierSetM = Ap . identifierSetM
