@@ -1,9 +1,10 @@
 {-|
   Copyright   :  (C) 2012-2016, University of Twente,
                      2016     , Myrtle Software Ltd,
-                     2017     , Google Inc.
+                     2017     , Google Inc.,
+                     2021-2023, QBayLogic B.V.
   License     :  BSD2 (see the file LICENSE)
-  Maintainer  :  Christiaan Baaij <christiaan.baaij@gmail.com>
+  Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
   Turn CoreHW terms into normalized CoreHW Terms
 -}
@@ -15,15 +16,13 @@
 
 module Clash.Normalize where
 
-import           Control.Concurrent.Supply        (Supply)
 import           Control.Exception                (throw)
 import qualified Control.Lens                     as Lens
-import           Control.Monad                    (when)
+import           Control.Monad                    ((>=>), when)
 import           Control.Monad.State.Strict       (State)
 import           Data.Default                     (def)
 import           Data.Either                      (lefts,partitionEithers)
 import qualified Data.IntMap                      as IntMap
-import           Data.IntMap.Strict               (IntMap)
 import           Data.List
   (intersect, mapAccumL)
 import qualified Data.Map                         as Map
@@ -37,31 +36,23 @@ import           Prettyprinter                    (vcat)
 import           Data.Text.Prettyprint.Doc        (vcat)
 #endif
 
-#if MIN_VERSION_ghc(9,0,0)
-import           GHC.Types.Basic                  (InlineSpec (..))
-#else
-import           BasicTypes                       (InlineSpec (..))
-#endif
+import           GHC.BasicTypes.Extra             (isNoInline)
 
 import           Clash.Annotations.BitRepresentation.Internal
   (CustomReprs)
-
-#if EXPERIMENTAL_EVALUATOR
-import           Clash.Core.PartialEval           (Evaluator)
-#else
-import           Clash.Core.Evaluator.Types       (Evaluator)
-#endif
-
+import           Clash.Core.Evaluator.Types as WHNF (Evaluator)
 import           Clash.Core.FreeVars
-  (freeLocalIds, globalIds, globalIdOccursIn, localIdDoesNotOccurIn)
+  (freeLocalIds, globalIds)
+import           Clash.Core.HasFreeVars           (notElemFreeVars)
+import           Clash.Core.HasType
+import           Clash.Core.PartialEval as PE     (Evaluator)
 import           Clash.Core.Pretty                (PrettyOptions(..), showPpr, showPpr', ppr)
 import           Clash.Core.Subst
   (extendGblSubstList, mkSubst, substTm)
 import           Clash.Core.Term                  (Term (..), collectArgsTicks
                                                   ,mkApps, mkTicks)
 import           Clash.Core.Type                  (Type, splitCoreFunForallTy)
-import           Clash.Core.TyCon
-  (TyConMap, TyConName)
+import           Clash.Core.TyCon (TyConMap)
 import           Clash.Core.Type                  (isPolyTy)
 import           Clash.Core.Var                   (Id, varName, varType)
 import           Clash.Core.VarEnv
@@ -70,7 +61,7 @@ import           Clash.Core.VarEnv
    mkVarEnv, mkVarSet, notElemVarEnv, notElemVarSet, nullVarEnv, unionVarEnv)
 import           Clash.Debug                      (traceIf)
 import           Clash.Driver.Types
-  (BindingMap, Binding(..), ClashOpts (..), DebugLevel (..))
+  (BindingMap, Binding(..), DebugOpts(..), ClashEnv(..))
 import           Clash.Netlist.Types
   (HWMap, FilteredHWType(..))
 import           Clash.Netlist.Util
@@ -79,15 +70,16 @@ import           Clash.Normalize.Strategy
 import           Clash.Normalize.Transformations
 import           Clash.Normalize.Types
 import           Clash.Normalize.Util
-import           Clash.Primitives.Types           (CompiledPrimMap)
-import           Clash.Rewrite.Combinators        ((>->),(!->),repeatR,topdownR)
+import           Clash.Rewrite.Combinators
+  ((>->), (!->), bottomupR, repeatR, topdownR)
 import           Clash.Rewrite.Types
-  (RewriteEnv (..), RewriteState (..), bindings, dbgLevel, dbgRewriteHistoryFile, extra,
-   tcCache, topEntities)
+  (RewriteEnv (..), RewriteState (..), bindings, debugOpts, extra,
+   tcCache, topEntities, newInlineStrategy)
 import           Clash.Rewrite.Util
   (apply, isUntranslatableType, runRewriteSession)
 import           Clash.Util
 import           Clash.Util.Interpolate           (i)
+import           Clash.Util.Supply                (Supply)
 
 import           Data.Binary                      (encode)
 import qualified Data.ByteString                  as BS
@@ -99,8 +91,7 @@ import           Clash.Rewrite.Types (RewriteStep(..))
 
 -- | Run a NormalizeSession in a given environment
 runNormalization
-  :: ClashOpts
-  -- ^ Level of debug messages to print
+  :: ClashEnv
   -> Supply
   -- ^ UniqueSupply
   -> BindingMap
@@ -108,52 +99,36 @@ runNormalization
   -> (CustomReprs -> TyConMap -> Type ->
       State HWMap (Maybe (Either String FilteredHWType)))
   -- ^ Hardcoded Type -> HWType translator
-  -> CustomReprs
-  -> TyConMap
-  -- ^ TyCon cache
-  -> IntMap TyConName
-  -- ^ Tuple TyCon cache
-  -> Evaluator
+  -> PE.Evaluator
   -- ^ Hardcoded evaluator for partial evaluation
-  -> CompiledPrimMap
-  -- ^ Primitive Definitions
+  -> WHNF.Evaluator
+  -- ^ Hardcoded evaluator for WHNF (old evaluator)
   -> VarEnv Bool
   -- ^ Map telling whether a components is part of a recursive group
   -> [Id]
   -- ^ topEntities
   -> NormalizeSession a
   -- ^ NormalizeSession to run
-  -> a
-runNormalization opts supply globals typeTrans reprs tcm tupTcm eval primMap rcsMap topEnts
-  = runRewriteSession rwEnv rwState
+  -> IO a
+runNormalization env supply globals typeTrans peEval eval rcsMap topEnts =
+  runRewriteSession rwEnv rwState
   where
+    -- TODO The RewriteEnv should just take ClashOpts.
     rwEnv     = RewriteEnv
-                  (opt_dbgLevel opts)
-                  (opt_dbgTransformations opts)
-                  (opt_dbgTransformationsFrom opts)
-                  (opt_dbgTransformationsLimit opts)
-                  (opt_dbgRewriteHistoryFile opts)
-                  (opt_aggressiveXOpt opts)
+                  env
                   typeTrans
-                  tcm
-                  tupTcm
+                  peEval
                   eval
                   (mkVarSet topEnts)
-                  reprs
-                  (opt_evaluatorFuelLimit opts)
 
     rwState   = RewriteState
                   0
+                  mempty       -- transformCounters Map
                   globals
                   supply
                   (error $ $(curLoc) ++ "Report as bug: no curFun",noSrcSpan)
                   0
-#if EXPERIMENTAL_EVALUATOR
-                  IntMap.empty
-                  0
-#else
                   (IntMap.empty, 0)
-#endif
                   emptyVarEnv
                   normState
 
@@ -161,38 +136,42 @@ runNormalization opts supply globals typeTrans reprs tcm tupTcm eval primMap rcs
                   emptyVarEnv
                   Map.empty
                   emptyVarEnv
-                  (opt_specLimit opts)
                   emptyVarEnv
-                  (opt_inlineLimit opts)
-                  (opt_inlineFunctionLimit opts)
-                  (opt_inlineConstantLimit opts)
-                  primMap
                   Map.empty
                   rcsMap
-                  (opt_newInlineStrat opts)
-                  (opt_ultra opts)
-                  (opt_inlineWFCacheLimit opts)
-
 
 normalize
   :: [Id]
   -> NormalizeSession BindingMap
-normalize []  = return emptyVarEnv
-normalize top = do
-  (new,topNormalized) <- unzip <$> mapM normalize' top
-  newNormalized <- normalize (concat new)
-  return (unionVarEnv (mkVarEnv topNormalized) newNormalized)
+normalize = go >=> unionWithCache
+ where
+  go []  = return emptyVarEnv
+  go top = do
+    (new,topNormalized) <- unzip <$> mapM normalize' top
+    newNormalized <- normalize (concat new)
+    return (unionVarEnv (mkVarEnv topNormalized) newNormalized)
+
+  unionWithCache :: BindingMap -> NormalizeSession BindingMap
+  unionWithCache env = do
+    cache <- Lens.use (extra.normalized)
+    -- We need to include the cache in our final result, forgetting to do so
+    -- leads to https://github.com/clash-lang/clash-compiler/issues/3109
+    --
+    -- On the other hand, just returning the cache as our final result could
+    -- not be enough, because normalize' might return a non-normalized binder
+    -- that is later picked up and cleaned up by flattenCallTree.
+    return (unionVarEnv cache env)
 
 normalize' :: Id -> NormalizeSession ([Id], (Id, Binding Term))
 normalize' nm = do
   exprM <- lookupVarEnv nm <$> Lens.use bindings
   let nmS = showPpr (varName nm)
   case exprM of
-    Just (Binding nm' sp inl pr tm) -> do
+    Just (Binding nm' sp inl pr tm r) -> do
       tcm <- Lens.view tcCache
       topEnts <- Lens.view topEntities
       let isTop = nm `elemVarSet` topEnts
-          ty0 = varType nm'
+          ty0 = coreTypeOf nm'
           ty1 = if isTop then tvSubstWithTyEq ty0 else ty0
 
       -- check for polymorphic types
@@ -215,11 +194,11 @@ normalize' nm = do
       resTyRep <- not <$> isUntranslatableType False resTy
       if resTyRep
          then do
-            tmNorm <- normalizeTopLvlBndr isTopEnt nm (Binding nm' sp inl pr tm)
+            tmNorm <- normalizeTopLvlBndr isTopEnt nm (Binding nm' sp inl pr tm r)
             let usedBndrs = Lens.toListOf globalIds (bindingTerm tmNorm)
-            traceIf (nm `elem` usedBndrs)
+            traceIf (bindingRecursive tmNorm)
                     (concat [ $(curLoc),"Expr belonging to bndr: ",nmS ," (:: "
-                            , showPpr (varType (bindingId tmNorm))
+                            , showPpr (coreTypeOf (bindingId tmNorm))
                             , ") remains recursive after normalization:\n"
                             , showPpr (bindingTerm tmNorm) ])
                     (return ())
@@ -242,13 +221,13 @@ normalize' nm = do
             -- for the ByteArray# inside of a Natural constant.
             -- (GHC-8.4 does this with tests/shouldwork/Numbers/Exp.hs)
             -- It will later be inlined by flattenCallTree.
-            lvl <- Lens.view dbgLevel
-            traceIf (lvl > DebugNone)
+            opts <- Lens.view debugOpts
+            traceIf (dbg_invariants opts)
                     (concat [$(curLoc), "Expr belonging to bndr: ", nmS, " (:: "
-                            , showPpr (varType nm')
+                            , showPpr (coreTypeOf nm')
                             , ") has a non-representable return type."
                             , " Not normalising:\n", showPpr tm] )
-                    (return ([],(nm,(Binding nm' sp inl pr tm))))
+                    (return ([],(nm,(Binding nm' sp inl pr tm r))))
 
 
     Nothing -> error $ $(curLoc) ++ "Expr belonging to bndr: " ++ nmS ++ " not found"
@@ -266,11 +245,8 @@ checkNonRecursive norm = case mapMaybeVarEnv go norm of
                                  | (a,b) <- eltsVarEnv rcs
                                  ])
  where
-  go (Binding nm _ _ _ tm) =
-    if nm `globalIdOccursIn` tm
-       then Just (nm,tm)
-       else Nothing
-
+  go (Binding nm _ _ _ tm r) =
+    if r then Just (nm,tm) else Nothing
 
 -- | Perform general \"clean up\" of the normalized (non-recursive) function
 -- hierarchy. This includes:
@@ -332,8 +308,8 @@ stripArgs _ _ _ = Nothing
 flattenNode
   :: CallTree
   -> NormalizeSession (Either CallTree ((Id,Term),[CallTree]))
-flattenNode c@(CLeaf (_,(Binding _ _ NoInline _ _))) = return (Left c)
-flattenNode c@(CLeaf (nm,(Binding _ _ _ _ e))) = do
+flattenNode c@(CLeaf (_,(Binding _ _ spec _ _ _))) | isNoInline spec = return (Left c)
+flattenNode c@(CLeaf (nm,(Binding _ _ _ _ e _))) = do
   isTopEntity <- elemVarSet nm <$> Lens.view topEntities
   if isTopEntity then return (Left c) else do
     tcm  <- Lens.view tcCache
@@ -342,13 +318,13 @@ flattenNode c@(CLeaf (nm,(Binding _ _ _ _ e))) = do
       Right (ids,[(bId,bExpr)],_) -> do
         let (fun,args,ticks) = collectArgsTicks bExpr
         case stripArgs ids (reverse ids) (reverse args) of
-          Just remainder | bId `localIdDoesNotOccurIn` bExpr ->
+          Just remainder | bId `notElemFreeVars` bExpr ->
                return (Right ((nm,mkApps (mkTicks fun ticks) (reverse remainder)),[]))
           _ -> return (Right ((nm,e),[]))
       _ -> return (Right ((nm,e),[]))
-flattenNode b@(CBranch (_,(Binding _ _ NoInline _ _)) _) =
+flattenNode b@(CBranch (_,(Binding _ _ spec _ _ _)) _) | isNoInline spec =
   return (Left b)
-flattenNode b@(CBranch (nm,(Binding _ _ _ _ e)) us) = do
+flattenNode b@(CBranch (nm,(Binding _ _ _ _ e _)) us) = do
   isTopEntity <- elemVarSet nm <$> Lens.view topEntities
   if isTopEntity then return (Left b) else do
     tcm  <- Lens.view tcCache
@@ -357,11 +333,11 @@ flattenNode b@(CBranch (nm,(Binding _ _ _ _ e)) us) = do
       Right (ids,[(bId,bExpr)],_) -> do
         let (fun,args,ticks) = collectArgsTicks bExpr
         case stripArgs ids (reverse ids) (reverse args) of
-          Just remainder | bId `localIdDoesNotOccurIn` bExpr ->
+          Just remainder | bId `notElemFreeVars` bExpr ->
                return (Right ((nm,mkApps (mkTicks fun ticks) (reverse remainder)),us))
           _ -> return (Right ((nm,e),us))
       _ -> do
-        newInlineStrat <- Lens.use (extra.newInlineStrategy)
+        newInlineStrat <- Lens.view newInlineStrategy
         if newInlineStrat || isCheapFunction e
            then return (Right ((nm,e),us))
            else return (Left b)
@@ -370,7 +346,7 @@ flattenCallTree
   :: CallTree
   -> NormalizeSession CallTree
 flattenCallTree c@(CLeaf _) = return c
-flattenCallTree (CBranch (nm,(Binding nm' sp inl pr tm)) used) = do
+flattenCallTree (CBranch (nm,(Binding nm' sp inl pr tm r)) used) = do
   flattenedUsed   <- mapM flattenCallTree used
   (newUsed,il_ct) <- partitionEithers <$> mapM flattenNode flattenedUsed
   let (toInline,il_used) = unzip il_ct
@@ -381,7 +357,8 @@ flattenCallTree (CBranch (nm,(Binding nm' sp inl pr tm)) used) = do
       let tm1 = substTm "flattenCallTree.flattenExpr" subst tm
 
       -- NB: When -fclash-debug-history is on, emit binary data holding the recorded rewrite steps
-      rewriteHistFile <- Lens.view dbgRewriteHistoryFile
+      opts <- Lens.view debugOpts
+      let rewriteHistFile = dbg_historyFile opts
       when (Maybe.isJust rewriteHistFile) $
         let !_ = unsafePerformIO
              $ BS.appendFile (Maybe.fromJust rewriteHistFile)
@@ -399,35 +376,36 @@ flattenCallTree (CBranch (nm,(Binding nm' sp inl pr tm)) used) = do
   -- inline all components when the resulting expression after flattening
   -- is still considered "cheap". This happens often at the topEntity which
   -- wraps another functions and has some selectors and data-constructors.
-  if inl /= NoInline && isCheapFunction newExpr
+  if not (isNoInline inl) && isCheapFunction newExpr
      then do
         let (toInline',allUsed') = unzip (map goCheap allUsed)
             subst' = extendGblSubstList (mkSubst emptyInScopeSet)
                                         (Maybe.catMaybes toInline')
         let tm1 = substTm "flattenCallTree.flattenCheap" subst' newExpr
         newExpr' <- rewriteExpr ("flattenCheap",flatten) (showPpr nm, tm1) (nm', sp)
-        return (CBranch (nm,(Binding nm' sp inl pr newExpr')) (concat allUsed'))
-     else return (CBranch (nm,(Binding nm' sp inl pr newExpr)) allUsed)
+        return (CBranch (nm,(Binding nm' sp inl pr newExpr' r)) (concat allUsed'))
+     else return (CBranch (nm,(Binding nm' sp inl pr newExpr r)) allUsed)
   where
     flatten =
-      repeatR (topdownR (apply "appPropFast" appPropFast >->
+      repeatR (topdownR (apply "appProp" appProp >->
                  apply "bindConstantVar" bindConstantVar >->
                  apply "caseCon" caseCon >->
-#if EXPERIMENTAL_EVALUATOR
-                 apply "deadcode" deadCode >->
-#else
                  (apply "reduceConst" reduceConst !-> apply "deadcode" deadCode) >->
-#endif
                  apply "reduceNonRepPrim" reduceNonRepPrim >->
-                 apply "removeUnusedExpr" removeUnusedExpr >->
-                 apply "flattenLet" flattenLet)) !->
-      topdownSucR (apply "topLet" topLet)
+                 apply "removeUnusedExpr" removeUnusedExpr) >->
+               bottomupR (apply "flattenLet" flattenLet)) !->
+      topdownSucR (apply "topLet" topLet) >->
+      -- See [Note] relation `collapseRHSNoops` and `inlineCleanup`
+      -- Note that we do this as the very last step, after all constant propagation
+      -- has been done to avoid #3036.
+      topdownSucR (apply "collapseRHSNoops" collapseRHSNoops) >->
+      topdownSucR (apply "inlineCleanup" inlineCleanup)
 
-    goCheap c@(CLeaf   (nm2,(Binding _ _ inl2 _ e)))
-      | inl2 == NoInline = (Nothing     ,[c])
+    goCheap c@(CLeaf   (nm2,(Binding _ _ inl2 _ e _)))
+      | isNoInline inl2  = (Nothing     ,[c])
       | otherwise        = (Just (nm2,e),[])
-    goCheap c@(CBranch (nm2,(Binding _ _ inl2 _ e)) us)
-      | inl2 == NoInline = (Nothing, [c])
+    goCheap c@(CBranch (nm2,(Binding _ _ inl2 _ e _)) us)
+      | isNoInline inl2  = (Nothing, [c])
       | otherwise        = (Just (nm2,e),us)
 
 callTreeToList :: [Id] -> CallTree -> ([Id], [(Id, Binding Term)])
