@@ -1,5 +1,5 @@
 {-|
-  Copyright   :  (C) 2020 QBayLogic
+  Copyright   :  (C) 2020,2022 QBayLogic B.V.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -27,7 +27,41 @@ import Data.Text.Lazy (pack)
 import System.IO (hPutStrLn, stderr)
 import Text.Trifecta.Result (Result(Success))
 
-#if MIN_VERSION_ghc(9,0,0)
+#if MIN_VERSION_ghc(9,8,0)
+import GHC.Unit.Module.Warnings (emptyWarningCategorySet)
+import GHC.Utils.Error
+  (DiagOpts(..), mkPlainDiagnostic, mkPlainMsgEnvelope, pprLocMsgEnvelopeDefault)
+import GHC.Utils.Outputable
+  (blankLine, empty, int, integer, showSDocUnsafe, text, ($$), ($+$), (<+>),
+   defaultSDocContext )
+import qualified GHC.Utils.Outputable as Outputable
+import GHC.Types.Error (DiagnosticReason (WarningWithoutFlag))
+import GHC.Types.SrcLoc (isGoodSrcSpan)
+#elif MIN_VERSION_ghc(9,6,0)
+import GHC.Utils.Error
+  (DiagOpts(..), mkPlainDiagnostic, mkPlainMsgEnvelope, pprLocMsgEnvelopeDefault)
+import GHC.Utils.Outputable
+  (blankLine, empty, int, integer, showSDocUnsafe, text, ($$), ($+$), (<+>),
+   defaultSDocContext )
+import qualified GHC.Utils.Outputable as Outputable
+import GHC.Types.Error (DiagnosticReason (WarningWithoutFlag))
+import GHC.Types.SrcLoc (isGoodSrcSpan)
+#elif MIN_VERSION_ghc(9,4,0)
+import GHC.Utils.Error
+  (DiagOpts(..), mkPlainDiagnostic, mkPlainMsgEnvelope, pprLocMsgEnvelope)
+import GHC.Utils.Outputable
+  (blankLine, empty, int, integer, showSDocUnsafe, text, ($$), ($+$), (<+>),
+   defaultSDocContext )
+import qualified GHC.Utils.Outputable as Outputable
+import GHC.Types.Error (DiagnosticReason (WarningWithoutFlag))
+import GHC.Types.SrcLoc (isGoodSrcSpan)
+#elif MIN_VERSION_ghc(9,2,0)
+import GHC.Utils.Error (mkPlainWarnMsg, pprLocMsgEnvelope)
+import GHC.Utils.Outputable
+  (blankLine, empty, int, integer, showSDocUnsafe, text, ($$), ($+$), (<+>))
+import qualified GHC.Utils.Outputable as Outputable
+import GHC.Types.SrcLoc (isGoodSrcSpan)
+#elif MIN_VERSION_ghc(9,0,0)
 import GHC.Driver.Session (unsafeGlobalDynFlags)
 import GHC.Utils.Error (mkPlainWarnMsg, pprLocErrMsg)
 import GHC.Utils.Outputable
@@ -70,16 +104,36 @@ toIntegerBB :: HDL -> HWType -> BlackBoxFunction
 toIntegerBB hdl hty _isD _primName args _ty = do
   case args of
     (Right (LitTy (NumTy i)):_) -> do
-      iw <- Lens.use intWidth
+      iw <- Lens.view intWidth
       let i1 = width i
       when (fromInteger i1 > iw) $ do
         (_,sp) <- Lens.use curCompNm
         let srcInfo1 | isGoodSrcSpan sp = srcInfo
                      | otherwise        = empty
-
+#if MIN_VERSION_ghc(9,8,0)
+            opts     = DiagOpts mempty mempty emptyWarningCategorySet emptyWarningCategorySet False False Nothing defaultSDocContext
+            diag     = mkPlainDiagnostic WarningWithoutFlag [] (warnMsg i1 iw $+$ blankLine $+$ srcInfo1)
+            warnMsg1 = mkPlainMsgEnvelope opts sp diag
+            warnMsg2 = pprLocMsgEnvelopeDefault warnMsg1
+#elif MIN_VERSION_ghc(9,6,0)
+            opts     = DiagOpts mempty mempty False False Nothing defaultSDocContext
+            diag     = mkPlainDiagnostic WarningWithoutFlag [] (warnMsg i1 iw $+$ blankLine $+$ srcInfo1)
+            warnMsg1 = mkPlainMsgEnvelope opts sp diag
+            warnMsg2 = pprLocMsgEnvelopeDefault warnMsg1
+#elif MIN_VERSION_ghc(9,4,0)
+            opts     = DiagOpts mempty mempty False False Nothing defaultSDocContext
+            diag     = mkPlainDiagnostic WarningWithoutFlag [] (warnMsg i1 iw $+$ blankLine $+$ srcInfo1)
+            warnMsg1 = mkPlainMsgEnvelope opts sp diag
+            warnMsg2 = pprLocMsgEnvelope warnMsg1
+#elif MIN_VERSION_ghc(9,2,0)
+            warnMsg1 = mkPlainWarnMsg sp (warnMsg i1 iw $+$ blankLine $+$ srcInfo1)
+            warnMsg2 = pprLocMsgEnvelope warnMsg1
+#else
             warnMsg1 = mkPlainWarnMsg unsafeGlobalDynFlags sp (warnMsg i1 iw $+$ blankLine $+$ srcInfo1)
+            warnMsg2 = pprLocErrMsg warnMsg1
+#endif
 
-        liftIO (hPutStrLn stderr (showSDocUnsafe (pprLocErrMsg warnMsg1)))
+        liftIO (hPutStrLn stderr (showSDocUnsafe warnMsg2))
     _ -> return ()
   return ((meta,) <$> bb)
  where
@@ -99,7 +153,7 @@ toIntegerBB hdl hty _isD _primName args _ty = do
     _ -> case hty of
       BitVector {} -> "~IF~SIZE[~TYP[1]]~THEN~IF~CMPLE[~SIZE[~TYPO]][~SIZE[~TYP[1]]]~THEN$unsigned(~VAR[bv][1][0+:~SIZE[~TYPO]])~ELSE$unsigned({{(~SIZE[~TYPO]-~SIZE[~TYP[1]]) {1'b0}},~VAR[bv][1]})~FI~ELSE~SIZE[~TYPO]'sd0~FI"
       Index {}     -> "~IF~SIZE[~TYP[0]]~THEN~IF~CMPLE[~SIZE[~TYPO]][~SIZE[~TYP[0]]]~THEN$unsigned(~VAR[i][0][0+:~SIZE[~TYPO]])~ELSE$unsigned({{(~SIZE[~TYPO]-~SIZE[~TYP[0]]) {1'b0}},~VAR[i][0]})~FI~ELSE~SIZE[~TYPO]'sd0~FI"
-      Signed {}    -> "~IF~SIZE[~TYP[0]]~THEN~IF~CMPLE[~SIZE[~TYPO]][~SIZE[~TYP[0]]]~THEN$signed(~VAR[i][0][0+:~SIZE[~TYPO]])~ELSE$signed({{(~SIZE[~TYPO]-~SIZE[~TYP[0]]) {1'b0}},~VAR[i][0]})~FI~ELSE~SIZE[~TYPO]'sd0~FI"
+      Signed {}    -> "~IF~SIZE[~TYP[0]]~THEN~IF~CMPLE[~SIZE[~TYPO]][~SIZE[~TYP[0]]]~THEN$signed(~VAR[i][0][0+:~SIZE[~TYPO]])~ELSE$signed({{(~SIZE[~TYPO]-~SIZE[~TYP[0]]) {~VAR[i][0][~SIZE[~TYP[0]]-1]}},~VAR[i][0]})~FI~ELSE~SIZE[~TYPO]'sd0~FI"
       Unsigned {}  -> "~IF~SIZE[~TYP[0]]~THEN~IF~CMPLE[~SIZE[~TYPO]][~SIZE[~TYP[0]]]~THEN$unsigned(~VAR[i][0][0+:~SIZE[~TYPO]])~ELSE$unsigned({{(~SIZE[~TYPO]-~SIZE[~TYP[0]]) {1'b0}},~VAR[i][0]})~FI~ELSE~SIZE[~TYPO]'sd0~FI"
       _            -> error "internal error"
 
