@@ -3,6 +3,7 @@
                     2017     , Myrtle Software Ltd
                     2018     , Google Inc.
                     2021     , QBayLogic B.V.
+                    2022     , Google Inc.
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -17,12 +18,14 @@ module Clash.Primitives.Util
   ( generatePrimMap
   , hashCompiledPrimMap
   , constantArgs
-  , decodeOrErr
+  , decodeOrErrJson
+  , decodeOrErrYaml
   , getFunctionPlurality
   ) where
 
 import           Control.DeepSeq        (force)
-import           Data.Aeson.Extra       (decodeOrErr)
+import           Control.Monad          (forM)
+import           Data.Aeson.Extra       (decodeOrErrJson, decodeOrErrYaml)
 import qualified Data.ByteString.Lazy   as LZ
 import qualified Data.HashMap.Lazy      as HashMap
 import qualified Data.HashMap.Strict    as HashMapStrict
@@ -58,8 +61,8 @@ import           Clash.Netlist.BlackBox.Types
 hashCompiledPrimitive :: CompiledPrimitive -> Int
 hashCompiledPrimitive (Primitive {name, primSort}) = hash (name, primSort)
 hashCompiledPrimitive (BlackBoxHaskell {function}) = fst function
-hashCompiledPrimitive (BlackBox {name, kind, outputReg, libraries, imports, includes, template}) =
-  hash (name, kind, outputReg, libraries, imports, includes', hashBlackbox template)
+hashCompiledPrimitive (BlackBox {name, kind, outputUsage, libraries, imports, includes, template}) =
+  hash (name, kind, outputUsage, libraries, imports, includes', hashBlackbox template)
     where
       includes' = map (\(nms, bb) -> (nms, hashBlackbox bb)) includes
       hashBlackbox (BBTemplate bbTemplate) = hash bbTemplate
@@ -96,7 +99,7 @@ resolvePrimitive' _metaPath (Primitive name wf primType) =
   return (name, HasBlackBox [] (Primitive name wf primType))
 resolvePrimitive' metaPath BlackBox{template=t, includes=i, resultNames=r, resultInits=ri, ..} = do
   let resolveSourceM = traverse (traverse (resolveTemplateSource metaPath))
-  bb <- BlackBox name workInfo renderVoid multiResult kind () outputReg libraries imports functionPlurality
+  bb <- BlackBox name workInfo renderVoid multiResult kind () outputUsage libraries imports functionPlurality
           <$> mapM (traverse resolveSourceM) i
           <*> traverse resolveSourceM r
           <*> traverse resolveSourceM ri
@@ -115,8 +118,13 @@ resolvePrimitive
   => FilePath
   -> IO [(TS.Text, GuardedResolvedPrimitive)]
 resolvePrimitive fileName = do
-  prims <- decodeOrErr fileName <$> LZ.readFile fileName
+  prims <- decoder fileName <$> LZ.readFile fileName
   mapM (resolvePrimitive' fileName) prims
+ where
+  decoder
+   | ".primitives.yaml" `isSuffixOf` fileName = decodeOrErrYaml
+   | ".primitives" `isSuffixOf` fileName = decodeOrErrJson
+   | otherwise = error ("Unexpected filename extension in: " <> fileName)
 
 addGuards
   :: ResolvedPrimMap
@@ -157,21 +165,23 @@ generatePrimMap
   -- ^ unresolved primitives found in annotations (in LoadModules and
   -- LoadInterfaceFiles)
   -> [(TS.Text, PrimitiveGuard ())]
+  -- ^ Primitive guards found in annotations
   -> [FilePath]
   -- ^ Directories to search for primitive definitions
   -> IO ResolvedPrimMap
 generatePrimMap unresolvedPrims primGuards filePaths = do
-  primitiveFiles <- fmap concat $ mapM
-     (\filePath -> do
-         fpExists <- Directory.doesDirectoryExist filePath
-         if fpExists
-           then
-             fmap ( map (FilePath.combine filePath)
-                  . filter (isSuffixOf ".primitives")
-                  ) (Directory.getDirectoryContents filePath)
-           else
-             return []
-     ) filePaths
+  primitiveFiles <- fmap concat $ forM filePaths $ \filePath -> do
+    fpExists <- Directory.doesDirectoryExist filePath
+    if fpExists then do
+      contents <- Directory.getDirectoryContents filePath
+      let
+        jsonPrims = filter (".primitives" `isSuffixOf`) contents
+        yamlPrims = filter (".primitives.yaml" `isSuffixOf`) contents
+        relPrims = jsonPrims <> yamlPrims
+        absPrims = map (FilePath.combine filePath) relPrims
+      return absPrims
+    else
+      return []
 
   primitives0 <- concat <$> mapM resolvePrimitive primitiveFiles
   let metapaths = map (TS.unpack . name) unresolvedPrims
