@@ -1,5 +1,5 @@
 {-|
-  Copyright  :  (C) 2020 QBayLogic B.V.
+  Copyright  :  (C) 2020,2022 QBayLogic B.V.
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -11,7 +11,7 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE QuasiQuotes #-}
-{-# LANGUAGE TemplateHaskell #-}
+{-# LANGUAGE TemplateHaskellQuotes #-}
 
 module Test.Clash.Rewrite where
 
@@ -22,15 +22,15 @@ import qualified Clash.Core.Literal as C
 import qualified Clash.Core.Type as C
 import qualified Clash.Core.Var as C
 import Clash.Core.VarEnv (InScopeSet, emptyVarSet, emptyVarEnv, emptyInScopeSet)
-import Clash.Driver.Types (DebugLevel(DebugSilent))
+import Clash.Driver.Types (ClashEnv(..), ClashOpts(..), defClashOpts, debugSilent)
 import Clash.Rewrite.Types
 import Clash.Rewrite.Util (runRewrite)
 import Clash.Normalize.Types
-import Clash.Unique (emptyUniqMap)
 import qualified Clash.Util.Interpolate as I
+import Clash.Util.Supply (newSupply)
+import Clash.Unique (Unique)
 
 import Control.Applicative ((<|>))
-import Control.Concurrent.Supply (newSupply)
 import Data.Default
 import Language.Haskell.Exts.Syntax
 import Language.Haskell.Exts.Parser (parseExp, fromParseResult)
@@ -44,13 +44,12 @@ import qualified Language.Haskell.TH.Quote as TH
 import qualified Data.List as List
 import qualified Data.HashMap.Strict as HashMap
 import qualified Data.Map as Map
-import qualified Data.Set as Set
 import qualified Data.IntMap as IntMap
 import qualified Data.Text as Text
 
-type TypeMap = HashMap.HashMap Int C.Type
+type TypeMap = HashMap.HashMap Unique C.Type
 
-lookupTM :: Int -> TypeMap -> C.Type
+lookupTM :: Unique -> TypeMap -> C.Type
 lookupTM u tm = case HashMap.lookup u tm of
   Just t -> t
   Nothing ->
@@ -61,35 +60,30 @@ lookupTM u tm = case HashMap.lookup u tm of
 
 instance Default RewriteEnv where
   def = RewriteEnv
-    { _dbgLevel=DebugSilent
-    , _dbgTransformations=Set.empty
-    , _dbgTransformationsFrom=0
-    , _dbgTransformationsLimit=maxBound
-    , _dbgRewriteHistoryFile=Nothing
-    , _aggressiveXOpt=False
+    { _clashEnv = ClashEnv
+        { envOpts = defClashOpts { opt_debug = debugSilent }
+        , envTyConMap = mempty
+        , envTupleTyCons = IntMap.empty
+        , envPrimitives = HashMap.empty
+        , envCustomReprs = buildCustomReprs []
+        , envDomains = HashMap.empty
+        }
     , _typeTranslator=error "_typeTranslator: NYI"
-    , _tcCache=emptyUniqMap
-    , _tupleTcCache=IntMap.empty
+    , _peEvaluator=error "_peEvaluator: NYI"
     , _evaluator=error "_evaluator: NYI"
     , _topEntities=emptyVarSet
-    , _customReprs=buildCustomReprs []
-    , _fuelLimit=10
     }
 
 instance Default extra => Default (RewriteState extra) where
   def = RewriteState
     { _transformCounter=0
+    , _transformCounters=mempty
     , _bindings=emptyVarEnv
     , _uniqSupply=unsafePerformIO newSupply
     , _curFun=error "_curFun: NYI"
     , _nameCounter=2
     , _workFreeBinders=emptyVarEnv
-#if EXPERIMENTAL_EVALUATOR
-    , _ioHeap=error "_ioHeap: NYI"
-    , _ioAddr=0
-#else
     , _globalHeap=error "_globalHeap: NYI"
-#endif
     , _extra=def
     }
 
@@ -98,17 +92,9 @@ instance Default NormalizeState where
     { _normalized=emptyVarEnv
     , _specialisationCache=Map.empty
     , _specialisationHistory=emptyVarEnv
-    , _specialisationLimit=20
     , _inlineHistory=emptyVarEnv
-    , _inlineLimit=20
-    , _inlineFunctionLimit=15
-    , _inlineConstantLimit=0
-    , _primitives=HashMap.empty
     , _primitiveArgs=Map.empty
     , _recursiveComponents=emptyVarEnv
-    , _newInlineStrategy=True
-    , _normalizeUltra=False
-    , _inlineWFCacheLimit=10
     }
 
 instance Default InScopeSet where
@@ -126,9 +112,10 @@ runSingleTransformation
   -- ^ Transformation to perform
   -> C.Term
   -- ^ Term to transform
-  -> C.Term
-runSingleTransformation rwEnv rwState is trans term = t
- where (t, _, _) = runR (runRewrite "" is trans term) rwEnv rwState
+  -> IO C.Term
+runSingleTransformation rwEnv rwState is trans term = do
+  (t, _, _) <- runR (runRewrite "" is trans term) rwEnv rwState
+  pure t
 
 -- | Run a single transformation with an empty environment and empty
 -- InScopeSet. See Default instances ^ to inspect the precise definition of
@@ -138,7 +125,7 @@ runSingleTransformation rwEnv rwState is trans term = t
 -- include a type translator, evaluator, current function, or global heap. Maps,
 -- like the primitive and tycon map, are also empty. If the transformation under
 -- test needs these definitions, you should add them manually.
-runSingleTransformationDef :: Default extra => Rewrite extra -> C.Term -> C.Term
+runSingleTransformationDef :: Default extra => Rewrite extra -> C.Term -> IO C.Term
 runSingleTransformationDef = runSingleTransformation def def def
 
 
@@ -231,7 +218,7 @@ parseDecls typs0 decls = (typs1, map parseOtherDecl otherDecls)
     e ->
       error ("parseOtherDecl: " <> show e)
 
-  parseTypeDecl :: Decl l -> [(Int, C.Type)]
+  parseTypeDecl :: Decl l -> [(Unique, C.Type)]
   parseTypeDecl (TypeSig _ nms t) =
     map (\nm -> (C.nameUniq (parseName nm), parseType t)) nms
   parseTypeDecl _ = error "impossible"
